
import numpy as np
from astropy.convolution import Gaussian2DKernel, convolve

def expand_interpolate(data, ker_sig=1):
    # data should be a 3D array
    kernel = Gaussian2DKernel(ker_sig)  # large kernel size because regions outside the guesses are likely noisy

    data_smooth = data.copy()

    if data.ndim == 2:
        data_smooth = convolve(data, kernel, boundary='extend')
        qmask = np.isnan(data)
        data[qmask] = data_smooth[qmask]

    elif data.ndim == 3:
        for i, gsm in enumerate(data_smooth):
            data_smooth[i] = convolve(data[i], kernel, boundary='extend')
            qmask = np.isnan(data[i])
            data[i][qmask] = data_smooth[i][qmask]

    return data


def iter_expand(data, mask):
    # expand the interpolation outwards, with increasingly larger kernals
    # until all pixels in the mask are filled

    def get_expand(data, mask):
        if data.ndim == 3:
            expand = np.sum(np.any(np.isnan(data), axis=0)[mask]) > 0
        elif data.ndim == 2:
            expand = np.sum(np.isnan(data)[mask]) > 0
        return expand

    expand = get_expand(data, mask)

    i = 0
    while expand:
        data = expand_interpolate(data, ker_sig= 1 +i)
        expand = get_expand(data, mask)
        i = i+ 1
    try:
        data[:, ~mask] = np.nan
    except IndexError:
        data[~mask] = np.nan

    return data


"""
The `mufasa.multi_v_fit` module provides functionality for preprocessing, masking, and refining
guesses for spectral line fitting, as well as utility functions for calculating
moments, signal-to-noise ratios, and handling parameter constraints.
"""

from __future__ import print_function
from __future__ import absolute_import

__author__ = 'mcychen'

import numpy as np
import pyspeckit
import astropy.io.fits as fits
import copy
import os, errno
import multiprocessing
import warnings
from datetime import datetime

from astropy import units as u
from astropy.units import UnitConversionError
from astropy.convolution import Gaussian2DKernel, convolve
from astropy.stats import mad_std

from scipy import ndimage as ndi

from spectral_cube import SpectralCube
from spectral_cube.lower_dimensional_structures import Projection

from pyspeckit.spectrum.units import SpectroscopicAxis

from skimage.morphology import remove_small_objects, disk, opening, binary_erosion, remove_small_holes, binary_dilation
from dask.array.core import Array as daskArray
import dask.array

from . import signals
from . import moment_guess as momgue
from . import guess_refine as g_refine
from .exceptions import SNRMaskError, FitTypeError, StartFitError
from .utils.multicore import validate_n_cores
from .utils import interpolate
from . import __version__
from .spec_models.meta_model import MetaModel

# =======================================================================================================================

from .utils.mufasa_log import get_logger

logger = get_logger(__name__)


# =======================================================================================================================

def get_chisq(cube, model, expand=20, reduced=True, usemask=True, mask=None, rest_value=None):

    if rest_value is not None:
        cube = cube.with_spectral_unit(u.Hz, rest_value=rest_value)

    if usemask:
        if mask is None:
            mask = model > 0
    else:
        mask = ~np.isnan(model)

    residual = cube.filled_data[:].value - model

    # This calculates chisq over the region where the fit is non-zero
    # plus a buffer of size set by the expand keyword.

    selem = np.ones(expand, dtype=np.bool)
    selem.shape += (1, 1,)
    mask = ndi.binary_dilation(mask, selem)
    mask = mask.astype(np.float)
    chisq = np.sum((residual * mask) ** 2, axis=0)

    if reduced:
        chisq /= np.sum(mask, axis=0)

    # This produces a robust estimate of the RMS along every line of sight:
    # (alternatively, we can use mad_std from astropy?)
    diff = residual - np.roll(residual, 2, axis=0)
    rms = 1.4826 * np.nanmedian(np.abs(diff), axis=0) / 2 ** 0.5

    chisq /= rms ** 2

    if reduced:
        # return the reduce chi-squares values
        return chisq

    else:
        # return the ch-squared values and the number of data points used
        return chisq, np.sum(mask, axis=0)


def register_pcube(pcube, mod_info):

    # unit check and conversion
    if hasattr(pcube, "unit"):
        if pcube.unit is None or pcube.unit == '':
            logger.warning("The cube doesn't have a unit. The fitting will proceed assume it has a unit of K")
            pcube.unit = 'K'
        elif pcube.unit != 'K':
            raise ValueError(f"The cube should have units of K rather than {pcube.unit}")
    else:
        logger.warning("The cube doesn't have a unit. The fitting will proceed assume it has a unit of K")

    # spectral unit check and conversion
    if pcube.xarr.refX is None or np.isnan(pcube.wcs.wcs.restfrq):
        # Specify the reference rest frequency if not present
        logger.warning("The cube has no reference rest frequency."
                       " The rest frequency of the spectral model will be used instead")
        pcube.xarr.refX = mod_info.freq_dict[linename] * u.Hz

    if pcube.xarr.velocity_convention is None:
        pcube.xarr.velocity_convention = 'radio'

    # always register the fitter just in case different lines are used
    fitter = mod_info.fitter
    pcube.specfit.Registry.add_fitter(mod_info.fittype, fitter, fitter.npars)
    logger.debug("number of parameters is {0}".format(fitter.npars))
    logger.debug("the line to fit is {0}".format(mod_info.linenames))
    return pcube

# setup_cube is depreciated
'''
def setup_cube(cube, mod_info, return_spectral_cube=False):

    if hasattr(cube, 'spectral_axis'):
        # if it's a SpectralCube object
        pcube = pyspeckit.Cube(cube=cube)

        if isinstance(cube._data, dask.array.Array):
            logger.info("SpectralCube uses dask")
            # load the entire array onto pcube
            pcube.data = cube._data.compute()
            pcube.data = cube._data.compute()


        if False:# #isinstance(cube._data, np.ndarray):
            logger.info("SpectralCube does not use dask")
            pcube = pyspeckit.Cube(cube=cube)

        elif False: # isinstance(cube._data, dask.array.Array):
            logger.info("SpectralCube uses dask")

            if cube.spectral_axis.flags['OWNDATA']:
                xarr = SpectroscopicAxis(cube.spectral_axis,
                                         unit=cube.spectral_axis.unit,
                                         refX=cube.wcs.wcs.restfrq,
                                         refX_unit='Hz')
            else:
                xarr = SpectroscopicAxis(cube.spectral_axis.copy(),
                                         unit=cube.spectral_axis.unit,
                                         refX=cube.wcs.wcs.restfrq,
                                         refX_unit='Hz')

            pcube = pyspeckit.Cube(cube=cube._data.compute(), xarr=xarr, header=cube.header)

            if (cube.unit in ('undefined', u.dimensionless_unscaled)
                    and 'BUNIT' in cube._meta):
                pcube.unit = cube._meta['BUNIT']
            else:
                pcube.unit = cube.unit

    elif isinstance(cube, str):
        cubename = cube
        if return_spectral_cube:
            cube = SpectralCube.read(cube)
            # pcube = pyspeckit.Cube(filename=cubename)
            pcube = pyspeckit.Cube(cube=cube)

        else:
            pcube = pyspeckit.Cube(filename=cubename)

    pcube.unit = "K"
    if np.isnan(pcube.wcs.wcs.restfrq):
        # Specify the rest frequency not present
        pcube.xarr.refX = mod_info.freq_dict[linename] * u.Hz
    pcube.xarr.velocity_convention = 'radio'

    # always register the fitter just in case different lines are used
    fitter = mod_info.fitter
    pcube.specfit.Registry.add_fitter(mod_info.fittype, fitter, fitter.npars)
    logger.debug("number of parameters is {0}".format(fitter.npars))
    logger.debug("the line to fit is {0}".format(mod_info.linenames))

    if return_spectral_cube:
        # the following check on rest-frequency may not be necessarily for GAS, but better be safe than sorry
        # note: this assume the data cube has the right units
        if np.isnan(cube._wcs.wcs.restfrq):
            # Specify the rest frequency not present
            cube = cube.with_spectral_unit(u.Hz, rest_value=mod_info.rest_value)
        cube = cube.with_spectral_unit(u.km / u.s, velocity_convention='radio')

        return pcube, cube
    else:
        return pcube
'''

def cubefit_simp(cube, pcube, ncomp, guesses, multicore=None, maskmap=None, fittype='nh3_multi_v', snr_min=None, **kwargs):
    # a simper version of cubefit_gen that assumes good user provided guesses

    logger.debug("using cubefit_simp")

    # get information on the spectral model
    mod_info = MetaModel(fittype, ncomp)
    freq_dict = mod_info.freq_dict
    linename = mod_info.linenames

    Texmin = mod_info.Texmin
    Texmax = mod_info.Texmax
    sigmin = mod_info.sigmin
    sigmax = mod_info.sigmax
    taumax = mod_info.taumax
    taumin = mod_info.taumin
    eps = mod_info.eps

    # deprecated and no longer needed
    #pcube = setup_cube(cube, mod_info, return_spectral_cube=False)

    register_pcube(pcube, mod_info)

    multicore = validate_n_cores(multicore)

    # get the masking for the fit
    footprint_mask = np.any(np.isfinite(cube._data), axis=0)
    planemask = np.any(np.isfinite(guesses), axis=0)

    peaksnr, planemask, kwargs = handle_snr(pcube, snr_min, planemask, **kwargs)

    if maskmap is not None:
        maskmap = maskmap & planemask & footprint_mask
    else:
        maskmap = planemask & footprint_mask

    logger.debug(f"maskmap type: {type(maskmap)}")
    maskmap = maskmap.compute()

    v_peak_hwidth = 10
    v_guess = guesses[::4]
    v_guess[v_guess == 0] = np.nan

    v_median, v_99p, v_1p = get_vstats(v_guess)

    vmax = v_median + v_peak_hwidth
    vmin = v_median - v_peak_hwidth

    # use percentile limits padded with sigmax if these values are more relaxed than the v_peak window
    if v_99p + sigmax > vmax:
        vmax = v_99p + sigmax
    if v_1p - sigmax < vmin:
        vmin = v_1p - sigmax

    def impose_lim(data, min=None, max=None, eps=0):
        if min is not None:
            mask = data < min
            data[mask] = min + eps
        if max is not None:
            mask = data > max
            data[mask] = max - eps

    # impose parameter limits on the guesses
    impose_lim(guesses[::4], vmin, vmax, eps)
    impose_lim(guesses[1::4], sigmin, sigmax, eps)
    impose_lim(guesses[2::4], Texmin, Texmax, eps)
    impose_lim(guesses[3::4], taumin, taumax, eps)

    kwargs['start_from_point'] = get_start_point(maskmap, weight=None)

    logger.info(f"Number of cores used for cubefit_simp: {multicore}")

    # add all the pcube.fiteach kwargs)
    kwargs['multicore'] = multicore
    kwargs['maskmap'] = maskmap
    kwargs['use_neighbor_as_guess'] = False
    kwargs['limitedmax'] = [True, True, True, True] * ncomp
    kwargs['maxpars'] = [vmax, sigmax, Texmax, taumax] * ncomp
    kwargs['limitedmin'] = [True, True, True, True] * ncomp
    kwargs['minpars'] = [vmin, sigmin, Texmin, taumin] * ncomp
    kwargs = set_pyspeckit_verbosity(**kwargs)

    pcube.fiteach(fittype=fittype, guesses=guesses, **kwargs)

    return pcube


def cubefit_gen(cube, pcube, ncomp=2, paraname=None, modname=None, chisqname=None, guesses=None, errmap11name=None,
                multicore=None, mask_function=None, snr_min=0.0, momedgetrim=True, saveguess=False,
                fittype='nh3_multi_v', **kwargs):

    logger.debug('Using cubefit_gen')

    # get information on the spectral model
    mod_info = MetaModel(fittype, ncomp)
    freq_dict = mod_info.freq_dict
    linename = mod_info.linenames

    Texmin = mod_info.Texmin
    Texmax = mod_info.Texmax
    sigmin = mod_info.sigmin
    sigmax = mod_info.sigmax
    taumax = mod_info.taumax
    taumin = mod_info.taumin
    eps = mod_info.eps

    register_pcube(pcube, mod_info)

    if errmap11name is not None:
        errmap = fits.getdata(errmap11name)
    else:
        # a quick way to estimate RMS as long as the noise dominates the spectrum by channels
        errmap = None

    linewidth_sigma = False
    trim = 3

    # trim the edges by 3 pixels
    cube = signals.trim_cube_edge(cube, trim=trim)
    m0, m1, m2, rms = signals.get_moments(cube, window_hwidth=mod_info.window_hwidth,
                                          linewidth_sigma=linewidth_sigma, trim=None,
                                          central_win_hwidth=mod_info.central_win_hwidth,
                                          return_rms=True)

    peaksnr = signals.get_snr(cube, rms=rms)

    # use min and max assuming the moment 1 is robust
    vmax = np.nanmax(m1).value + sigmax / 2
    vmin = np.nanmin(m1).value - sigmax / 2

    logger.debug("velocity limit: ({:.3f}, {:.3f})".format(vmin, vmax))

    if 'planemask' in kwargs:
        planemask = kwargs['planemask']
        planemask = planemask & np.any(cube.mask.include(), axis=0)
    else:
        planemask = np.any(cube.mask.include(), axis=0)

    if 'maskmap' in kwargs:
        logger.debug("including user specified mask as a base")
        planemask = np.logical_and(kwargs['maskmap'], planemask)

    if 'signal_cut' in kwargs:
        logger.warning("signal_cut will be set to zero for pcube.fiteach() in cubefit_gen."
                       " The snr_min is used to determine which pixels to fit prior to calling fiteach()")

    if snr_min > 0:
        snr_mask = peaksnr > snr_min
        if snr_min >= 3:
            mom_mask = np.isfinite(m1)
            planemask = planemask & snr_mask & mom_mask
        else:
            planemask = planemask & snr_mask

    if planemask.sum() < 1:
        msg = "The provided snr_min={} results in no valid pixels to fit; fitting terminated".format(snr_min)
        raise SNRMaskError(msg)

    if mask_function is not None:
        msg = "\'mask_function\' is now deprecation, and will be removed in the next version"
        warnings.warn(msg, DeprecationWarning)
        logger.warning(msg)

    # interpolate the moment maps
    mmm = interpolate.iter_expand(np.array([m0, m1, m2]), mask=planemask)
    m0, m1, m2 = mmm[0], mmm[1], mmm[2]

    logger.info("velocity fitting limits: ({}, {})".format(np.round(vmin, 2), np.round(vmax, 2)))

    # get the guesses based on moment maps
    # tex and tau guesses are chosen to reflect low density, diffusive gas that are likley to have low SNR
    gg = momgue.moment_guesses(m1, m2, ncomp, sigmin=sigmin, moment0=m0, linetype=mod_info.linetype,
                               mom0_floor=None)

    if guesses is None:
        guesses = gg

    else:
        # fill in the blanks with convolved interpolation then moment guesses
        guesses[guesses == 0] = np.nan
        gmask = np.isfinite(guesses)

        mom_mask = np.isfinite(gg)

        # get interpolated results
        kernel = Gaussian2DKernel(5)  # large kernel size because regions outside the guesses are likely noisy
        guesses_smooth = guesses.copy()
        for i, gsm in enumerate(guesses_smooth):
            guesses_smooth[i] = convolve(guesses[i], kernel, boundary='extend')

        guesses[~gmask] = guesses_smooth[~gmask]
        guesses[~mom_mask] = np.nan
        gmask = np.isfinite(guesses)

        # fill in the rest of the guesses with moment guesses
        guesses[~gmask] = gg[~gmask]

        # fill in the failed sigma guesses with interpotaed guesseses
        gmask = guesses[1::4] < sigmin
        guesses[1::4][gmask] = guesses_smooth[1::4][gmask]

        logger.debug("provided guesses accepted")

    # re-normalize the degenerated tau & text for the purpose of estimate guesses
    guesses[3::4], guesses[2::4] = g_refine.tautex_renorm(guesses[3::4], guesses[2::4], tau_thresh=taumin,
                                                          nu=mod_info.rest_value.to("GHz").value)

    # The guesses should be fine in the first case, but just in case, make sure the guesses are confined within the
    # appropriate limits
    guesses[::4][guesses[::4] > vmax] = vmax
    guesses[::4][guesses[::4] < vmin] = vmin
    guesses[1::4][guesses[1::4] > sigmax] = sigmax
    guesses[1::4][guesses[1::4] < sigmin] = sigmin + eps
    guesses[2::4][guesses[2::4] > Texmax] = Texmax
    guesses[2::4][guesses[2::4] < Texmin] = Texmin
    guesses[3::4][guesses[3::4] > taumax] = taumax
    guesses[3::4][guesses[3::4] < taumin] = taumin

    if saveguess:
        save_guesses(pcube, paraname, guesses, savename)

    kwargs['maskmap'] = planemask & np.any(cube.mask.include(), axis=0) & np.all(np.isfinite(guesses), axis=0)
    kwargs['maskmap'] = kwargs['maskmap'] & match_pcube_mask(pcube, kwargs['maskmap'])
    if isinstance(kwargs['maskmap'], daskArray):
        kwargs['maskmap'] = kwargs['maskmap'].compute()

    mask_size = np.sum(kwargs['maskmap'])

    if mask_size < 1:
        logger.warning("maskmap has no pixels, no fitting will be performed")
        return pcube
    elif np.all(np.isfinite(guesses), axis=0).sum() < 1:
        logger.warning("guesses has no pixels, no fitting will be performed")
        return pcube

    logger.info("Final mask size for the fitting: {0} pixels".format(mask_size))

    multicore = validate_n_cores(multicore)
    logger.info(f"Number of cores used for cubefit_gen: {multicore}")

    # get the start point
    if "start_from_point" in kwargs:
        logger.warning("user start point will not be used")
    start_point = get_start_point(kwargs['maskmap'], weight=peaksnr)
    kwargs['start_from_point'] = start_point

    # finalize the argruments to be passed into fiteach
    # signal_cut is set to zero, since a signal cut has already been done
    kwargs_core = dict(fittype=fittype,
                       guesses=guesses,
                       limitedmax=[True, True, True, True] * ncomp,
                       maxpars=[vmax, sigmax, Texmax, taumax] * ncomp,
                       limitedmin=[True, True, True, True] * ncomp,
                       minpars=[vmin, sigmin, Texmin, taumin] * ncomp,
                       multicore=multicore,
                       use_neighbor_as_guess=False,
                       guesses_are_moments=False,
                       signal_cut=0)
    kwargs = {**kwargs, **kwargs_core}
    kwargs = set_pyspeckit_verbosity(**kwargs)

    # start fitting the cube
    try:
        # fitting the cube
        idx = kwargs['start_from_point']
        logger.debug("guesses at start, 1st try: {}".format(guesses[:, idx[1], idx[0]]))
        logger.debug("peaksnr at start, 1st try: {}".format(peaksnr[idx[1], idx[0]]))
        pcube.fiteach(**kwargs)
    except ValueError as e:
        msg = "The starting fit position is somehow not among the valid, likely due to a bug. " \
              "ValueError message from pyspeckit: " + e.__str__()
        logger.warning(msg)
        retry_fit(pcube, kwargs, start_point, peaksnr=peaksnr)
    except AssertionError as e:
        msg = "The starting fit position is somehow not among the valid, likely due to a bug. " \
              "AssertionError message from pyspeckit: " + e.__str__()
        logger.warning(msg)
        retry_fit(pcube, kwargs, start_point, peaksnr=peaksnr)

    if paraname != None:
        save_pcube(pcube, paraname, ncomp=ncomp)

    if modname != None:
        model = SpectralCube(pcube.get_modelcube(), pcube.wcs, header=cube.header)
        model.write(modname, overwrite=True)

    if chisqname != None:
        chisq = get_chisq(cube, pcube.get_modelcube(), expand=20, rest_value=mod_info.rest_value)
        chisqfile = fits.PrimaryHDU(data=chisq, header=cube.wcs.celestial.to_header())
        chisqfile.writeto(chisqname, overwrite=True)

    return pcube


# ===============================================================================

def default_masking(snr, snr_min=5.0):
    # a snr masking fucntion that perform some quailty controls

    if snr_min is None:
        planemask = np.isfinite(snr)
    else:
        planemask = (snr > snr_min)

    if planemask.sum() > 100:
        # to create a less noisy mask further
        planemask = remove_small_holes(planemask, area_threshold=9)
        planemask = remove_small_objects(planemask, min_size=25)
        planemask = opening(planemask, disk(1))

    return (planemask)


def match_pcube_mask(pcube, maskmap):
    # a function to ensure the mask passed into pcube.fiteach will be valid
    # using the same code as pcube.fiteach to return a maskmap that should be valid when passed into fiteach
    if not hasattr(pcube.mapplot, 'plane'):
        logger.debug("pcube.mapplot has no attribute plane. Creating one right now.")
        pcube.mapplot.makeplane()
    if isinstance(pcube.mapplot.plane, np.ma.core.MaskedArray):
        OK = ((~pcube.mapplot.plane.mask) &
              maskmap.astype('bool')).astype('bool')
    else:
        if isinstance(pcube.mapplot.plane, daskArray):
            pcube.mapplot.plane = pcube.mapplot.plane.compute()
        OK = (np.isfinite(pcube.mapplot.plane) &
              maskmap.astype('bool')).astype('bool')
    return OK


def retry_fit(pcube, kwargs, old_start_point, peaksnr=None):
    # a module to retry a fit with a different start point
    # remove the last failed pixel from the mask before trying again

    kwargs['maskmap'][old_start_point[1], old_start_point[0]] = False
    kwargs['start_from_point'] = get_start_point(kwargs['maskmap'], weight=None)
    idx = kwargs['start_from_point']
    logger.debug("guesses at start, 2nd try: {}".format(kwargs['guesses'][:, idx[1], idx[0]]))
    if peaksnr is not None:
        logger.debug("peaksnr (not used) at start, 2st try: {}".format(peaksnr[idx[1], idx[0]]))

    try:
        pcube.fiteach(**kwargs)
    except ValueError as e:
        msg = "The first two attempts to fit a pixel did not yield a fit. Likely due to a lack of signal or poor guesses. " \
              "ValueError message from pyspeckit: " + e.__str__()
        logger.error(msg)
        raise StartFitError(msg)
    except AssertionError as e:
        msg = "The first two attempts to fit a pixel did not yield a fit. Likely due to a lack of signal or poor guesses. " \
              "AssertionError message from pyspeckit: " + e.__str__()
        logger.error(msg)
        raise StartFitError(msg)


def make_header(ndim, ref_header):

    # initilizing a new, empthy header
    hdr_new = fits.Header()

    hdr_new.set(keyword='DATE', value=datetime.now().strftime('%Y-%m-%dT%H:%M:%S'), comment=f"Written by MUFASA v{__version__}")

    if ndim==3:
        keylist = ['SIMPLE', 'BITPIX', 'NAXIS', 'NAXIS1', 'NAXIS2', 'NAXIS3', 'OBJECT',
                   'BMAJ', 'BMIN', 'BPA', 'WCSAXES', 'CRPIX1', 'CRPIX2', 'CRPIX3',
                   'CDELT1', 'CDELT2', 'CDELT3', 'CUNIT1', 'CUNIT2', 'CUNIT3',
                   'CTYPE1', 'CTYPE2', 'CTYPE3', 'CRVAL1', 'CRVAL2', 'CRVAL3',
                   'LONPOLE', 'LATPOLE', 'WCSNAME', 'MJDREF', 'RADESYS',
                   'EQUINOX', 'SPECSYS']
    elif ndim==2:
        keylist = ['SIMPLE', 'BITPIX', 'NAXIS', 'NAXIS1', 'NAXIS2', 'OBJECT',
                   'BMAJ', 'BMIN', 'BPA', 'WCSAXES', 'CRPIX1', 'CRPIX2',
                   'CDELT1', 'CDELT2', 'CUNIT1', 'CUNIT2',
                   'CTYPE1', 'CTYPE2', 'CRVAL1', 'CRVAL2',
                   'LONPOLE', 'LATPOLE', 'WCSNAME', 'MJDREF',
                   'RADESYS', 'EQUINOX', 'SPECSYS']

    for key in keylist:
        # copy cards from the reference header if they exist
        if key in ref_header:
            hdr_new.append(copy.copy(ref_header.cards[key]))

    # make sure the dimension of the header is written correctly
    comments_dict = dict(NAXIS='Number of array dimensions', WCSAXES='Number of coordinate axes')
    for key, comment in comments_dict.items():
        if (key in hdr_new) and (hdr_new[key] != ndim):
            hdr_new.set(keyword=key, value=ndim, comment=comment)

    hdr_new.add_history("====================================================================")
    hdr_new.add_history(f"Written by MUFASA v{__version__} on {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}")
    hdr_new.add_comment("====================================================================")
    hdr_new.add_comment("MUFASA citation: Chen et al. (2020), DOI: 10.3847/1538-4357/ab7378")
    hdr_new.add_comment("MUFASA is available on GitHub at: https://github.com/mcyc/mufasa")

    return hdr_new


def save_pcube(pcube, savename, ncomp, npara=4, header_note=None):
    # a method to save the fitted parameter cube with relavent header information

    # create a new header using the pcube's cube header as a template
    hdr_new = make_header(ndim=3, ref_header=pcube.header)

    if header_note is None:
        header_note = f'Parameter maps derived from {ncomp}-comp model fits'

    hdr_new.set(keyword='NOTES', value=header_note, comment=None, before='DATE')

    # modify the units of the plane accordingly
    hdr_new.set(keyword ='CDELT3', value=1, comment='[unitless] Coordinate increment at reference point')
    hdr_new.set(keyword ='CTYPE3', value='FITPAR', comment='fitted parameters')
    hdr_new.set(keyword ='CRVAL3', value=0, comment='[unitless] Coordinate value at reference point')
    hdr_new.set(keyword='CRPIX3', value=1, comment='[unitless] Pixel coordinate of reference point')
    hdr_new.set(keyword ='CUNIT3', value='None', comment='[unitless] Units of coordinate increment and value')

    if ncomp == 1:
        hdr_new.set(keyword ='PLANE0', value= 'VELOCITY', comment='[km/s] vlsr')
        hdr_new.set(keyword ='PLANE1', value= 'SIGMA', comment='[km/s] velocity dispersion')
        hdr_new.set(keyword ='PLANE2', value  = 'TEX', comment='[K] excitation temperature')
        hdr_new.set(keyword ='PLANE3', value  = 'TAU', comment='[unitless] peak optical depth')

        hdr_new.set(keyword ='PLANE4', value  = 'eVELOCITY', comment='[km/s] estimated error of vlsr')
        hdr_new.set(keyword ='PLANE5', value  = 'eSIGMA', comment='[km/s] estimated error velocity dispersion')
        hdr_new.set(keyword ='PLANE6', value = 'eTEX', comment='[K] estimated error excitation temperature')
        hdr_new.set(keyword ='PLANE7', value = 'eTAU', comment='[unitless] estimated error of peak optical depth')

    elif ncomp > 1:
        for i in range(0, ncomp):
            hdr_new.set(keyword=f'PLANE{i * npara + 0}', value=f'VELOCITY_{i + 1}',
                        comment=f'[km/s] vlsr of component {i+1}')

            hdr_new.set(keyword=f'PLANE{i * npara + 1}', value=f'SIGMA_{i + 1}',
                        comment=f'[km/s] velocity dispersion of component {i + 1}')

            hdr_new.set(keyword=f'PLANE{i * npara + 2}', value=f'TEX_{i + 1}',
                        comment=f'[K] excitation temperature of component {i + 1}')

            hdr_new.set(keyword=f'PLANE{i * npara + 3}', value=f'TAU_{i + 1}',
                        comment=f'[unitless] peak optical depth of component {i + 1}')

        # the loop is split into two so the numbers will be written in ascending order
        for i in range(0, ncomp):
            hdr_new.set(keyword=f'PLANE{(ncomp + i) * npara + 0}', value=f'eVELOCITY_{i + 1}',
                        comment=f'[km/s] estimated error of component {i+1} vlsr')

            hdr_new.set(keyword=f'PLANE{(ncomp + i) * npara + 1}', value=f'eSIGMA_{i + 1}',
                        comment=f'[km/s] estimated error of component {i + 1} velocity dispersion')

            hdr_new.set(keyword=f'PLANE{(ncomp + i) * npara + 2}', value=f'eTEX_{i + 1}',
                        comment=f'[K] estimated error of component {i + 1} excitation temperature')

            hdr_new.set(keyword=f'PLANE{(ncomp + i) * npara + 3}', value=f'eTAU_{i + 1}',
                        comment=f'[unitless] estimated error of component {i + 1} peak optical depth')

    fitcubefile = fits.PrimaryHDU(data=np.concatenate([pcube.parcube, pcube.errcube]), header=hdr_new)
    fitcubefile.writeto(savename, overwrite=True)
    logger.debug("{} saved.".format(savename))


def save_guesses(pcube, paraname, guesses, savename):
    # a module to save guesses for diagnostic purposes

    # create a new header using the pcube's cube header as a template
    hdr_new = make_header(ndim=3, ref_header=pcube.header)

    # modify the units of the plane accordingly
    hdr_new.set(keyword ='CDELT3', value=1, comment='[unitless] Coordinate increment at reference point')
    hdr_new.set(keyword ='CTYPE3', value='FITPAR', comment='fitted parameters')
    hdr_new.set(keyword ='CRVAL3', value=0, comment='[unitless] Coordinate value at reference point')
    hdr_new.set(keyword='CRPIX3', value=1, comment='[unitless] Pixel coordinate of reference point')
    hdr_new.set(keyword ='CUNIT3', value='None', comment='[unitless] Units of coordinate increment and value')

    savedir = "{0}/{1}".format(os.path.dirname(paraname), "guesses")

    try:
        os.makedirs(savedir)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    savename = "{0}_guesses.fits".format(os.path.splitext(paraname)[0], "parameter_maps")
    savename = "{0}/{1}".format(savedir, os.path.basename(savename))

    fitcubefile = fits.PrimaryHDU(data=guesses, header=hdr_new)
    fitcubefile.writeto(savename, overwrite=True)


##################

def set_pyspeckit_verbosity(**kwargs):
    if 'verbose_level' not in kwargs:
        from .utils.mufasa_log import DEBUG, INFO
        if logger.getEffectiveLevel() < DEBUG:
            kwargs['verbose_level'] = 4
        elif logger.getEffectiveLevel() <= INFO:
            kwargs['verbose_level'] = 1
        else:
            kwargs['verbose_level'] = 0

    if (kwargs['verbose_level'] == 0) and ('verbose' not in kwargs):
        kwargs['verbose'] = False
        # pyspeckit defaults to verbose=True
        # if verbose=True and verbose_level=0 it prints output

    return kwargs


##################
def get_vstats(velocities, signal_mask=None):
    m1 = velocities
    mask = np.isfinite(m1)
    if signal_mask is not None:
        mask = np.logical_and(mask, signal_mask)
    m1_good = m1[mask]

    v_median = np.median(m1_good)
    v_99p = np.percentile(m1_good, 99)
    v_1p = np.percentile(m1_good, 1)
    return v_median, v_99p, v_1p


def get_start_point(maskmap, weight=None, return_xy=True):
    # if return_xy, return index in the order of xy (pcube.fiteach's convention) instead of yx
    # start_from_point is ordered in yx before it was returned upstream
    if weight is not None:
        if isinstance(weight, Projection):
            weight = weight.value
        # use the max position in weight within the mask as a start point
        weight = np.nan_to_num(weight * maskmap)
        try:
            start_from_point = np.unravel_index(weight.argmax(), weight.shape)  # get start_from_point in the yx order
            logger.debug("starting point obtained from using the weight: {}".format(start_from_point))
            logger.debug("value of maskmap at startpoint: {}".format(maskmap[start_from_point]))
            if not maskmap[start_from_point]:
                msg = "The start point is not within maskmap. This is a bug that needs to be addressed."
                logger.error(msg)
                raise IndexError(msg)
        except IndexError:
            logger.debug("Trouble getting good start point with weight. Trying again without the weight.")
            indx_g = np.argwhere(maskmap)
            start_from_point = (indx_g[0, 0], indx_g[0, 1])
    else:
        indx_g = np.argwhere(maskmap)
        start_from_point = (indx_g[0, 0], indx_g[0, 1])
    logger.debug("starting point: {}".format(start_from_point))
    if return_xy:
        return (start_from_point[1], start_from_point[0])
    else:
        return start_from_point


def snr_estimate(pcube, errmap, smooth=True):
    if hasattr(pcube, 'cube'):
        cube = pcube.cube
    else:
        # assume pcube is just a "regular" cube
        cube = pcube

    if errmap is None:
        # a quick way to estimate RMS as long as the noise dominates the spectrum by channels
        errmap = mad_std(cube, axis=0, ignore_nan=True)

    err_med = np.nanmedian(errmap)
    logger.debug("median rms: {:.5f}".format(err_med))

    # mask out pixels that are too noisy (in this case, 3 times the median rms in the cube)
    err_mask = errmap < err_med * 3.0

    snr = cube / errmap
    peaksnr = np.nanmax(snr, axis=0)

    if smooth:
        # the snr map can be noisy, so a little smoothing
        kernel = Gaussian2DKernel(1)
        peaksnr = convolve(peaksnr, kernel)

    return peaksnr, snr, errmap, err_mask


def handle_snr(pcube, snr_min, planemask, return_errmask=False, **kwargs):
    if snr_min is None:
        if 'signal_cut' not in kwargs:
            snr_min = 0.0
        else:
            snr_min = kwargs['signal_cut']
    elif 'signal_cut' in kwargs:
        if kwargs['signal_cut'] != snr_min:
            if kwargs['signal_cut'] > snr_min:
                snr_min = kwargs['signal_cut']
            else:
                kwargs['signal_cut'] = snr_min
            logger.warning("snr_min and signal_cut provided do not match."
                           " The higher value of the two, {}, is adopted".format(snr_min))
    else:
        kwargs['signal_cut'] = 0

    if 'errmap' not in kwargs:
        errmap = None

    peaksnr, snr, emap, err_mask = snr_estimate(pcube, errmap, smooth=True)

    if snr_min > 0:
        snr_mask = peaksnr > snr_min
        planemask = np.logical_and(planemask, snr_mask)

    if planemask.sum() < 1:
        msg = "The provided snr_min={} results in no valid pixels to fit; fitting terminated.".format(snr_min)
        raise SNRMaskError(msg)

    if return_errmask:
        return peaksnr, planemask, kwargs, err_mask
    else:
        return peaksnr, planemask, kwargs


from __future__ import print_function
__author__ = 'mcychen'

#=======================================================================================================================

import numpy as np

from pyspeckit.spectrum.models import model
from pyspeckit.spectrum.models.ammonia_constants import (line_names, freq_dict, voff_lines_dict, tau_wts_dict)
from pyspeckit.spectrum.models.ammonia_constants import (ckms, h, kb)

#=======================================================================================================================
# set CMB as a global constant

TCMB = 2.7315 # K

#=======================================================================================================================

def nh3_multi_v_model_generator(n_comp, linenames = None):
    """
    Works for up to 2 componet fits at the moment
    Parameters
    ----------
    n_comp : int
        The number of velocity componets to fit
    linenames : list
        A list of line names from the set ('oneone', ..., 'eighteight'); default is just 'oneone'
    Returns
    -------
    model : `model.SpectralModel`
        A SpectralModel class build from N different metastable inversion
        hyperfine models
    """
    n_para = n_comp*4
    idx_comp = np.arange(n_comp)

    if linenames is None:
        linenames = ['oneone']

    nlines = len(linenames)

    if nlines > 1:
        raise Exception("modeling more than one line yet to be implemented. Please only use one line for the time being")
        return None

    def nh3_vtau_multimodel(xarr, *args):
        # the parameters are in the order of vel, width, tex, tau for each velocity component
        assert len(args) == n_para
        return ammonia_multi_v(xarr, *args, line_names=linenames)

    mod = model.SpectralModel(nh3_vtau_multimodel, n_para,
            parnames=[x
                      for ln in idx_comp
                      for x in ('vlsr{0}'.format(ln),
                                'sigma{0}'.format(ln),
                                'tex{0}'.format(ln),
                                'tau{0}'.format(ln))],
            parlimited=[(False,False), (True,False), (True,False), (True,False)]*n_para,
            parlimits=[(0,0), ]*n_para,
            shortvarnames=[x
                           for ln in idx_comp
                           for x in ('v_{{VLSR,{0}}}'.format(ln),
                                     '\\sigma_{{{0}}}'.format(ln),
                                     'T_{{ex,{0}}}'.format(ln),
                                     '\\tau_{{{0}}}'.format(ln))],
            fitunit='Hz')
            # the keyword fitunits is now fitunit?
    return mod


def ammonia_multi_v(xarr, *args, **kwargs):
    # the parameters are in the order of vel, width, tex, tau for each velocity component

    # note: it may be worth while replacing *args with a single list like keyword called parameters
    #args = parameters

    if xarr.unit.to_string() != 'GHz':
        xarr = xarr.as_unit('GHz')

    line_names = ["oneone"]

    if kwargs is not None:
        if "line_names" in kwargs:
            line_names = kwargs["line_names"]
            if len(line_names) > 1:
                raise Exception("modeling more than one line yet to be implemented. Please only use one line for the time being")
                return None

    background_ta = T_antenna(TCMB, xarr.value)
    tau_dict = {}

    # iteratively move through each slabs towards the observer (i.e., radiative transfer)
    for vel, width, tex, tau in zip(args[::4], args[1::4], args[2::4], args[3::4]):

        for linename in line_names:
            tau_dict[linename] = tau

        model_spectrum = _ammonia_spectrum(xarr, tex=tex, tau_dict=tau_dict, width=width, xoff_v=vel,
                                           line_names=line_names, background_ta=background_ta)

        # Set the spectrum of the current component as the background of the subsequent component
        background_ta = model_spectrum

    return model_spectrum - T_antenna(TCMB, xarr.value)


def _ammonia_spectrum(xarr, tex, tau_dict, width, xoff_v, line_names, background_ta=0.0, fillingfraction=None,
                      return_components=False):
    """
    Helper function: given a dictionary of ammonia optical depths, an excitation tmeperature... etc, and produce a
    spectrum based on a one-slab (i.e., single velocity component model)
    Note: this is a modified version of the _ammonia_spectrum found in pyspeckit/spectrum/models/ammonia.py
    Note2: the final spectrum returned do not have the background emission subtracted

    Parameters
    ----------
    background_ta : float or ndarray
        "Antenna temperature" of the background emission
    Returns
    -------
    model : `model.SpectralModel`
        A SpectralModel class build from N different metastable inversion
        hyperfine models

    """
    width_min = 0.01

    # fillingfraction is an arbitrary scaling for the data; the model will be (normal model) * fillingfraction
    if fillingfraction is None:
        fillingfraction = 1.0

    # "runspec" means "running spectrum": it is accumulated over a loop
    runspec = np.zeros(len(xarr))

    if return_components:
        components = []

    for linename in line_names:
        voff_lines = np.array(voff_lines_dict[linename])
        tau_wts = np.array(tau_wts_dict[linename])

        lines = (1-voff_lines/ckms)*freq_dict[linename]/1e9
        tau_wts = tau_wts / (tau_wts).sum()
        if width == 0: width = width_min  # width shouldn't be 0, saves from divide by 0
        nuwidth = np.abs(width/ckms*lines)
        nuoff = xoff_v/ckms*lines

        # tau array
        tauprof = np.zeros(len(xarr))
        for kk,nuo in enumerate(nuoff):
            tauprof_ = (tau_dict[linename] * tau_wts[kk] *
                        np.exp(-(xarr.value+nuo-lines[kk])**2 /
                               (2.0*nuwidth[kk]**2)))
            if return_components:
                components.append(tauprof_)
            tauprof += tauprof_

        T0 = (h*xarr.value*1e9/kb) # "temperature" of wavelength

        if isinstance(tex, dict):
            runspec = ((T0/(np.exp(T0/tex[linename])-1)*(1-np.exp(-tauprof)) +
                        background_ta*np.exp(-tauprof))*fillingfraction
                       + runspec)

        else:
            if tex < TCMB: 
                tex = TCMB # tex shouldn't be < TCMB unless we're looking at absorption?, also removes a divide by zero warning
            runspec = ((T0/(np.exp(T0/tex)-1)*(1-np.exp(-tauprof)) +
                        background_ta*np.exp(-tauprof))*fillingfraction
                       + runspec)

    if return_components:
        components = np.array(components)

        if isinstance(tex, dict):
            term1 = [(T0/(np.exp(T0/tex[linename])-1)*(1-np.exp(-1*components)) + background_ta*np.exp(-1*components))
                     for linename in line_names]
        else:
            term1 = (T0/(np.exp(T0/tex)-1)*(1-np.exp(-1*components)) + background_ta*np.exp(-1*components))

        # May want to make sure I include fillingfraction here in the future
        return term1
    else:
        return runspec


def T_antenna(Tbright, nu):
    """
    Calculate antenna temperatures over nu (in GHz)
    """
    T0 = (h*nu*1e9/kb)
    return T0/(np.exp(T0/Tbright)-1)




from multiprocessing import cpu_count

from .mufasa_log import get_logger

def validate_n_cores(n_cores):
    default = cpu_count() - 1

    if n_cores is None: return default

    if isinstance(n_cores, bool):
        if n_cores: return default
        else: return 1

    if isinstance(n_cores, int):
        if 1 <= n_cores <= cpu_count(): return n_cores
        else: raise ValueError(f'n_cores should be between 1 and multiprocessing.cpu_count(). Value given was {n_cores}')

    raise ValueError(f'n_cores must be one of (None, True, False, integer). Value given was {n_cores}')



"""
The `mufasa.clean_fits` module provides functionality for handling multi-component fitting
results, cleaning parameter maps, and generating exclusive single- and
two-component maps.
"""
from __future__ import print_function
__author__ = 'mcychen'

import numpy as np
from astropy.io import fits
#=======================================================================================================================
from .utils.mufasa_log import get_logger
logger = get_logger(__name__)
#=======================================================================================================================

class fit_results(object):

    def __init__(self, path_para_1c, path_lnk01):

        self.paraCubes = {}
        self.headers = {}
        self.lnkMaps = {}

        self.paraCubes['1c'], self.headers['1c'] = fits.getdata(path_para_1c, header=True)
        self.lnkMaps['10'] = fits.getdata(path_lnk01)
        self.ncompList = [1]


    def add_results(self, path_para, path_lnk, ncomp, path_lnkn0=None):
        if ncomp > 1:
            self.paraCubes['{}c'.format(ncomp)] = fits.getdata(path_para)
            self.headers['{}c'.format(ncomp)] = fits.getheader(path_para)
            self.lnkMaps['{}{}'.format(ncomp, ncomp-1)] = fits.getdata(path_lnk)
            if path_lnkn0 is not None:
                self.lnkMaps['{}{}'.format(ncomp, 0)] = fits.getdata(path_lnkn0)
            self.ncompList.append(ncomp)
        else:
            logger.error("ncomp must be >1. The provide value is {}. No action taken.".format(ncomp))

#=======================================================================================================================

def clean_2comp_maps(fit_results, savename=None, vErrThresh=None, removeExtremeV=True):
    fr=fit_results
    pmaps_1c = fr.paraCubes['1c'].copy()
    pmaps_2c = fr.paraCubes['2c'].copy()
    lnk21 = fr.lnkMaps['21']
    lnk10 =  fr.lnkMaps['10']

    remove_zeros(pmaps_1c)
    remove_zeros(pmaps_2c)

    # remove pixels better modeled by noise
    mask = lnk10 < 5
    pmaps_1c[:, mask] = np.nan

    # remove pixels best modeled by one component (or noise if lnk20 is provided)
    mask = lnk21 < 5
    if "20" in fr.lnkMaps:
        mask = np.logical_or(mask, fr.lnkMaps['20'] < 5)
    pmaps_2c[:, mask] = np.nan


    if removeExtremeV:
        # remove pixels with fitted vlsr & sigma that are 'stuck' at the extreme values
        mask_exmv_1c = extremeV_mask(pmaps_1c)
        mask_exmv_2c = extremeV_mask(pmaps_2c)

        pmaps_1c[:, mask_exmv_1c] = np.nan
        pmaps_2c[:, mask_exmv_2c] = np.nan

    if vErrThresh is not None:
        # remove pixels with fitted vlsr & sigma above the specified thresholds
        mask_thr_1c = above_ErrV_Thresh(pmaps_1c, vErrThresh)
        mask_thr_2c = above_ErrV_Thresh(pmaps_2c, vErrThresh)

        pmaps_1c[:, mask_thr_1c] = np.nan
        pmaps_2c[:, mask_thr_2c] = np.nan

    # replace two comp fit with 1 comp fit over where 2 comp pixels are empty
    mmask = ~np.all(np.isfinite(pmaps_2c), axis=0)
    pmaps_2c[0:4, mmask] = pmaps_1c[0:4, mmask]  # fitted parameters
    pmaps_2c[8:12, mmask] = pmaps_1c[4:8, mmask]  # fitted errors


    if savename is not None:
        fits.writeto(savename, data=pmaps_2c, header=fr.headers['2c'], overwrite=True)

    return pmaps_2c

#=======================================================================================================================


def exclusive_2comp_maps(clean_maps):#, hdr1, hdr2, path_1c, path_2c):
    # take the clean map and save them into files that exclusively only have the best fitted one comp or two comp maps

    # create empty arrays
    pmaps_2cx = np.empty(clean_maps.shape)
    pmaps_2cx[:] = np.nan
    pmaps_1cx = pmaps_2cx[0:8].copy()

    mask_2c = np.all(np.isfinite(clean_maps), axis=0)

    pmaps_1cx[0:4, ~mask_2c] = clean_maps[0:4, ~mask_2c]  # fitted parameters
    pmaps_1cx[4:8, ~mask_2c] = clean_maps[8:12, ~mask_2c]  # fitted errors

    pmaps_2cx[:, mask_2c] = clean_maps[:, mask_2c]  # fitted errors
    return pmaps_1cx, pmaps_2cx


def remove_zeros(pmaps):
    pmaps[pmaps==0] = np.nan

#=======================================================================================================================
# masking functions


def extremeV_mask(pmaps):
    # find pixels with extreme fitted vlsr or sigma values

    ncomp = pmaps.shape[0]/8

    # create a list of indices for all vlsr and sigma maps
    iList = []
    for i in range(ncomp):
        iList.append(i*4)     # add vlsr
        iList.append(i*4+1)   # add sigma

    # create an empty mask
    mmask = np.zeros(pmaps[0].shape, dtype='bool')
    for i in iList:
        mask1 = pmaps[i] == np.nanmin(pmaps[i])
        #print(np.sum(mask1))
        mask2 = pmaps[i] == np.nanmax(pmaps[i])
        #print(np.sum(mask2))
        mask = np.logical_or(mask1, mask2)
        mmask = np.logical_or(mmask, mask)

    return mask


def above_ErrV_Thresh(pmaps, thresh):
    # return a mask indicating the pixels with vlsr and sigma errors above the threshold

    ncomp = pmaps.shape[0]/8

    # create a list of indices for all vlsr and sigma maps
    iList = []
    for i in range(ncomp):
        iList.append(4*(i+ncomp))     # add vlsr error
        iList.append(4*(i+ncomp)+1)   # add sigma error

    # create an empty mask
    mmask = np.zeros(pmaps[0].shape, dtype='bool')
    for i in iList:
        mmask = np.logical_or(mmask, pmaps[i] > thresh)

    return mmask


"""
The `mufasa.UltraCube` module provides tools for processing, analyzing, and visualizing spectral
cubes, particularly for fitting multi-component spectral models.
"""

from __future__ import print_function
from __future__ import absolute_import
__author__ = 'mcychen'

#======================================================================================================================#
import os
import warnings
from functools import wraps
import numpy as np

from spectral_cube import SpectralCube
# prevent any spectral-cube related warnings from being displayed.
from spectral_cube.utils import SpectralCubeWarning
warnings.filterwarnings(action='ignore', category=SpectralCubeWarning, append=True)
from copy import copy

import pyspeckit
import gc
from astropy import units as u
from astropy.units import UnitConversionError
import scipy.ndimage as nd

import dask.array as da

from . import aic
from . import multi_v_fit as mvf
from . import convolve_tools as cnvtool
from .spec_models import meta_model
from .utils.multicore import validate_n_cores
from .visualization.spec_viz import Plotter
#======================================================================================================================#
from .utils.mufasa_log import get_logger
logger = get_logger(__name__)
#======================================================================================================================#

class UltraCube(object):
    """
    A framework for multi-component spectral cube analysis and model fitting.

    The `UltraCube` class manages spectral cubes, supports multi-component model
    fitting, and provides tools for statistical evaluation, visualization, and
    residual analysis. It is designed to handle large-scale spectral datasets
    efficiently.

    Parameters
    ----------
    cubefile : str, optional
        Path to the FITS cube file. If not provided, a `SpectralCube` instance must be
        supplied via the `cube` parameter.
    cube : SpectralCube, optional
        A `SpectralCube` instance representing the input spectral data. This parameter
        is used only if `cubefile` is not provided.
    fittype : str, optional
        Type of spectral model to fit. Must be compatible with the `MetaModel` framework.
    snr_min : float, optional
        Minimum signal-to-noise ratio for considering a voxel during fitting.
    rmsfile : str, optional
        Path to the file containing RMS values for the cube.
    cnv_factor : int, optional, default=2
        Factor by which to spatially convolve the cube for pre-processing.
    n_cores : int or bool, optional, default=True
        Number of CPU cores to use for parallel processing. If `True`, uses all available
        cores minus one.

    Notes
    -----
    - The `UltraCube` class is designed to handle cubes with varying spatial
      and spectral resolutions. It uses the `pyspeckit` library for fitting
      and the `MetaModel` framework for defining custom spectral models.
    - Convolution, masking, and statistical evaluation tools are provided for
      pre- and post-processing of data.
    - If a file path is provided via `cubefile`, the cube is loaded using
      `SpectralCube.read`.
    - The convolution factor (`cnv_factor`) is applied to the spatial axes to
      reduce resolution before processing.
    - The number of CPU cores (`n_cores`) defaults to all but one available core
      for parallel computations, but can be explicitly set.

    """

    def __init__(self, cubefile=None, cube=None, fittype=None, snr_min=None, rmsfile=None, cnv_factor=2, n_cores=True):
        # to hold pyspeckit cubes for fitting
        self.pcubes = {}
        self.residual_cubes = {}
        self.rms_maps = {}
        self.Tpeak_maps = {}
        self.chisq_maps = {}
        self.rchisq_maps = {}
        self.rss_maps = {}
        self.NSamp_maps = {}
        self.AICc_maps = {}
        self.master_model_mask = None
        self.snr_min = 0.0
        self.cnv_factor = cnv_factor
        self.n_cores = validate_n_cores(n_cores)
        self.fittype = fittype
        self.plotter = None
        self.meta_model = None

        if cubefile is not None:
            self.cubefile = cubefile
            self.load_cube(cubefile)
        else:
            if hasattr(cube, 'spectral_axis'):
                # Load from a SpectralCube instance
                self.cube = cube

        if fittype is not None:
            # for the current usage, setting ncomp=1 is fine. Changes to MetalModel in the future will be needed
            # to elimiate needing ncomp during its intialization
            self.meta_model = meta_model.MetaModel(fittype=fittype, ncomp=1)

        if not snr_min is None:
            self.snr_min = snr_min

        if not rmsfile is None:
            self.rmsfile = rmsfile


    def make_header2D(self):
        """
        Create a 2D header by removing the specified axis.

        Parameters
        ----------
        header : astropy.io.fits.Header
            FITS header to modify.
        axis : int
            Axis to remove from the header.

        Returns
        -------
        astropy.io.fits.Header
            Modified 2D header.
        """
        return mvf.make_header(ndim=2, ref_header=self.cube.header)

    def load_cube(self, fitsfile):
        """
        Load a spectral cube from a FITS file and convert its units to K and km/s.

        The function ensures the cube has intensity units in Kelvin and a spectral axis
        in velocity units (km/s). If the cube lacks a rest frequency, it assigns one
        based on the spectral model.

        Parameters
        ----------
        fitsfile : str
            Path to the FITS cube file.

        Returns
        -------
        None
        """
        cube = SpectralCube.read(fitsfile, use_dask=True)
        cube = to_K(cube)

        if cube.spectral_axis.unit.is_equivalent(u.Hz):
            # assign rest frequency from the model before spectral axis velocity conversion
            if not hasattr(cube.wcs.wcs, 'restfrq') or np.isnan(cube.wcs.wcs.restfrq):
                logger.warning("The cube has no reference rest frequency. "
                               "The rest frequency of the spectral model will be used instead")
                cube = cube.with_spectral_unit(u.Hz, rest_value=mod_info.rest_value)

        self.cube = cube.with_spectral_unit(u.km / u.s, velocity_convention='radio')


    def load_pcube(self, pcube_ref=None):
        """
        Load a cube into a pyspeckit.SpectralCube object from a .fits file.

        Parameters
        ----------
        pcube_ref : pyspeckit.SpectralCube
            A pyspeckit.SpectralCube object that works with the same data cube. If provided, the new pcube's cube attribute
            will be pointed towards this reference cube to avoid reduandance and save memory

        Returns
        -------
        pcube : pyspeckit.SpectralCube
            The pyspeckit.SpectralCube object needed work with the fitting
        """
        # read the cube first as a SpectralCube object to performe unit conversion
        # Note: dask is set to False to ensure it's compitable with some of pyspeckit's functions
        # this loading mehod is not memory efficient, but needed workaround at the moment
        cube_temp = SpectralCube.read(self.cubefile, dask=False)
        cube_temp = to_K(cube_temp) # convert the unit to K;

        pcube = pyspeckit.Cube(cube=cube_temp)

        # premptively release memory
        del cube_temp
        gc.collect()

        if pcube_ref is not None:
            if is_K(pcube_ref.unit):
                # set pointer of the new pcube's cube to the reference one to remove redundancy and conserve memory
                pcube.cube = pcube_ref.cube

        if pcube.xarr.refX is None or np.isnan(pcube.wcs.wcs.restfrq):
            # Specify the reference rest frequency if not present
            logger.warning("The cube has no reference rest frequency."
                           " The rest frequency of the spectral model will be used instead")
            pcube.xarr.refX = mod_info.freq_dict[linename] * u.Hz

        if pcube.xarr.velocity_convention is None:
            pcube.xarr.velocity_convention = 'radio'

        return pcube



    def convolve_cube(self, savename=None, factor=None, edgetrim_width=5):
        """
        Convolve the SpectralCube to a lower spatial resolution by a specified factor.

        Parameters
        ----------
        savename : str, optional
            Path to save the convolved cube. If None, the convolved cube is not saved.
        factor : int, optional
            Factor by which to spatially convolve the cube. If None, the default `self.cnv_factor` is used.
        edgetrim_width : int, optional
            Number of pixels to trim at the edges after convolution. Default is 5.

        Returns
        -------
        None

        Notes
        -----
        - Convolution reduces the spatial resolution by the specified factor.
        - The resulting convolved cube is stored in `self.cube_cnv`.
        - This method uses the `convolve_sky_byfactor` function to perform spatial convolution.

        Raises
        ------
        ValueError
            If the cube cannot be convolved due to incompatible dimensions or data types.
        """
        if factor is None:
            factor = self.cnv_factor
        self.cube_cnv = convolve_sky_byfactor(self.cube, factor, savename, edgetrim_width=edgetrim_width)


    def get_cnv_cube(self, filename=None):
        """
        Load the convolved cube if the file exists, or create one if it does not.

        Parameters
        ----------
        filename : str, optional
            Path to the convolved cube file. If None, a new convolved cube is created using the default `cnv_factor`.

        Returns
        -------
        None
        """
        if filename is None:
            self.convolve_cube(factor=self.cnv_factor)
        elif os.path.exists(filename):
            self.cube_cnv = SpectralCube.read(filename, use_dask=True)
        else:
            logger.warning("The specified convolved cube file does not exist.")


    def fit_cube(self, ncomp, simpfit=False, **kwargs):
        """
        Fit the spectral cube with the specified number of components.

        Parameters
        ----------
        ncomp : int or list of int
            Number of components for the model. If a list is provided, fits are performed for each component.
        simpfit : bool, optional
            Whether to use a simplified fitting method (`cubefit_simp`) instead of the general fitting method (`cubefit_gen`).
        **kwargs : dict, optional
            Additional keyword arguments passed to `pyspeckit.Cube.fiteach`. Includes options like `multicore` and `snr_min`.

        Returns
        -------
        None

        Notes
        -----
        - The `multicore` parameter in kwargs controls parallel processing. By default, the number of cores set in `self.n_cores` is used.
        - Fit results are stored in `self.pcubes` for each component.

        Raises
        ------
        TypeError
            If `ncomp` is not an integer or a list of integers.
        """
        if not 'multicore' in kwargs:
            kwargs['multicore'] = self.n_cores

        if not 'snr_min' in kwargs:
            kwargs['snr_min'] = self.snr_min
        try:
            from collections import Iterable
        except ImportError:
            # for backwards compatibility
            from collections.abc import Iterable
        if not isinstance(ncomp, Iterable):
            ncomp = [ncomp]

        for nc in ncomp:
            # initiate pcube objects for fitting and model handling
            if self.pcubes:
                # use a pre-existing pcube as reference
                _, pcube_ref = next(iter(self.pcubes.items()))
                self.pcubes[str(nc)] = self.load_pcube(pcube_ref=pcube_ref)
            else:
                self.pcubes[str(nc)] = self.load_pcube()
            self.pcubes[str(nc)] = fit_cube(self.cube, self.pcubes[str(nc)], fittype=self.fittype, simpfit=simpfit, ncomp=nc, **kwargs)

            if self.pcubes[str(nc)].has_fit.sum() > 0 and hasattr(self.pcubes[str(nc)],'parcube'):
                # update model mask if any fit has been performed
                mod_mask = self.pcubes[str(nc)].get_modelcube(multicore=kwargs['multicore']) > 0
                self.include_model_mask(mod_mask)
            gc.collect()

    def has_fit(self, ncomp):
        """
        Return a mask indicating which pixels have been fitted with a model.

        Parameters
        ----------
        ncomp : int
            The number of components for the model being checked. This value is used to identify the modelled
            parameter cube (`parcube`) associated with the number of components.

        Returns
        -------
        np.ndarray
            A 2D boolean array with the same spatial dimensions as the data, where `True` indicates pixels
            that have been fitted (i.e., contain non-zero, finite values in the parameter cube for the specified
            model component) and `False` indicates pixels that were not fitted.
        """
        parcube = self.pcubes[str(ncomp)].parcube
        mask = np.any(parcube != 0, axis=0)
        mask = mask & np.any(np.isfinite(parcube), axis=0)
        return mask


    def include_model_mask(self, mask):
        # update the mask that shows were all the models are non-zero

        if self.master_model_mask is None:
            self.master_model_mask = mask
        else:
            self.master_model_mask = np.logical_or(self.master_model_mask, mask)

    def reset_model_mask(self, ncomps, multicore=True):
        #reset and re-generate master_model_mask for all the components in ncomps
        self.master_model_mask = None

        for nc in ncomps:
            if nc > 0 and hasattr(self.pcubes[str(nc)],'parcube'):
                # update model mask if any fit has been performed
                mod_mask = self.pcubes[str(nc)]._modelcube > 0
                self.include_model_mask(mod_mask)
            gc.collect()


    def save_fit(self, savename, ncomp, header_note=None):
        # note, this implementation currently relies on
        if hasattr(self.pcubes[str(ncomp)], 'parcube'):
            save_fit(self.pcubes[str(ncomp)], savename, ncomp, header_note=header_note)
        else:
            logger.warning("no fit was performed and thus no file will be saved")


    def load_model_fit(self, filename, ncomp, calc_model=True, multicore=None):
        self.pcubes[str(ncomp)] = load_model_fit(self.cubefile, filename, ncomp, self.fittype)
        if calc_model:
            if multicore is None: multicore = self.n_cores
            # update model mask
            mod_mask = self.pcubes[str(ncomp)].get_modelcube(multicore=multicore) > 0
            logger.debug("{}comp model mask size: {}".format(ncomp, np.sum(mod_mask)) )
            gc.collect()
            self.include_model_mask(mod_mask)

    def get_residual(self, ncomp, multicore=None):
        """
        Calculate the residual cube by subtracting the fitted model from the data.

        Parameters
        ----------
        ncomp : int
            Number of components used in the fitted model.
        multicore : int or bool, optional
            Number of cores to use for computation. Defaults to `self.n_cores`.

        Returns
        -------
        np.ndarray or dask.array.Array
            Residual array representing the difference between the data and the model fit.

        Notes
        -----
        - Residual cubes are stored in `self.residual_cubes` for reuse.
        - Residuals are computed for pixels with valid model fits.

        Raises
        ------
        ValueError
            If no model fit is available for the specified number of components.
        """
        if multicore is None: multicore = self.n_cores
        compID = str(ncomp)
        model = self.pcubes[compID].get_modelcube(multicore=multicore)
        self.residual_cubes[compID] = get_residual(self.cube, model)
        gc.collect()
        return self.residual_cubes[compID]


    def get_rms(self, ncomp):
        compID = str(ncomp)
        if not compID in self.residual_cubes:
            self.get_residual(ncomp)

        self.rms_maps[compID] = get_rms(self.residual_cubes[compID])
        return self.rms_maps[compID]


    def get_rss(self, ncomp, mask=None, planemask=None, update=True, expand=20):
        # residual sum of squares
        if mask is None:
            mask = self.master_model_mask
        # note: a mechanism is needed to make sure NSamp is consistient across the models
        rrs, nsamp = calc_rss(self, ncomp, usemask=True, mask=mask, return_size=True, update_cube=False,
                              planemask=planemask, expand=expand)

        # only include pixels with samples
        mask = nsamp < 1
        nsamp[mask] = np.nan
        # only if rss value is valid
        mask = np.logical_or(mask, rrs <= 0)
        rrs[mask] = np.nan
        if planemask is None:
            self.rss_maps[str(ncomp)] = rrs
            self.NSamp_maps[str(ncomp)] = nsamp
        else:
            self.rss_maps[str(ncomp)][planemask] = rrs
            self.NSamp_maps[str(ncomp)][planemask] = nsamp

    def get_Tpeak(self, ncomp):
        compID = str(ncomp)
        model = self.pcubes[compID].get_modelcube(multicore=self.n_cores)
        self.Tpeak_maps[compID] = get_Tpeak(model)
        return self.Tpeak_maps[compID]

    def get_chisq(self, ncomp, mask=None):
        if mask is None:
            mask = self.master_model_mask
        # note: a mechanism is needed to make sure NSamp is consistient across
        self.chisq_maps[str(ncomp)], self.NSamp_maps[str(ncomp)] = \
            calc_chisq(self, ncomp, reduced=False, usemask=True, mask=mask)

    def get_reduced_chisq(self, ncomp):
        # no mask is passed insnr_mask, and thus is not meant for model comparision
        compID = str(ncomp)
        self.rchisq_maps[compID]= \
            calc_chisq(self, ncomp, reduced=True, usemask=True, mask=None)
        return self.rchisq_maps[compID]


    def get_AICc(self, ncomp, update=False, planemask=None, expand=20, **kwargs):
        # recalculate AICc fresh if update is True
        compID = str(ncomp)
        if update or not compID in self.AICc_maps:
            # start the calculation fresh
            # note that zero component is assumed to have no free-parameter (i.e., no fitting)
            p = ncomp * 4
            self.get_rss(ncomp, update=update, planemask=planemask, expand=expand, **kwargs)
            if planemask is None or not compID in self.AICc_maps:
                self.AICc_maps[compID] = aic.AICc(rss=self.rss_maps[compID], p=p, N=self.NSamp_maps[compID])
            else:
                self.AICc_maps[compID][planemask] = aic.AICc(rss=self.rss_maps[compID][planemask], p=p, N=self.NSamp_maps[compID][planemask])
        return self.AICc_maps[compID]


    def get_AICc_likelihood(self, ncomp1, ncomp2, **kwargs):
        return calc_AICc_likelihood(self, ncomp1, ncomp2, **kwargs)

    def get_all_lnk_maps(self, ncomp_max=2, rest_model_mask=True, multicore=True):
        return get_all_lnk_maps(self, ncomp_max=ncomp_max, rest_model_mask=rest_model_mask, multicore=multicore)

    def get_best_2c_parcube(self, multicore=True, lnk21_thres=5, lnk20_thres=5, lnk10_thres=5, return_lnks=True):
        kwargs = dict(multicore=multicore, lnk21_thres=lnk21_thres, lnk20_thres=lnk20_thres,
                      lnk10_thres=lnk10_thres, return_lnks=return_lnks)
        return get_best_2c_parcube(self, **kwargs)

    def get_best_residual(self, cubetype=None):
        return None

    def get_plotter(self, update=False, spec_unit='km/s', **kwargs):
        """
        Initialize or update the Plotter instance for visualizing fitted spectra.

        Parameters
        ----------
        update : bool, optional
            If True, update the existing plotter instance (default is False).
        spec_unit : str, optional
            The spectral unit to use for plotting the spectral axis (default is 'km/s').
        **kwargs : dict, optional
            Additional keyword arguments passed to the `Plotter` class for initialization.

        Returns
        -------
        None
        """
        if self.plotter is None or update:
            self.plotter = Plotter(self, fittype=self.fittype, spec_unit=spec_unit, **kwargs)

    def plot_spec(self, x, y, ax=None, xlab=None, ylab=None, **kwargs):
        """
        Plot a single spectrum at (x, y).

        Parameters
        ----------
        x : int
            X-coordinate of the pixel.
        y : int
            Y-coordinate of the pixel.
        ax : matplotlib.axes.Axes, optional
            Axes object to plot on. If None, a new figure and axes are created.
        xlab : str, optional
            X-axis label. Default is the LSR velocity label.
        ylab : str, optional
            Y-axis label. Default is the main beam temperature label.
        **kwargs : dict
            Additional keyword arguments passed to `plot_spec`.

        Returns
        -------
        fig : matplotlib.figure.Figure
            The figure object.
        ax : matplotlib.axes.Axes
            The axes object.
        """
        self.get_plotter()
        return self.plotter.plot_spec(x, y, ax=ax, xlab=xlab, ylab=ylab, **kwargs)

    def plot_spec_grid(self, x, y, size=3, xsize=None, ysize=None, xlim=None, ylim=None, figsize=None, **kwargs):
        """
        Plot a grid of spectra centered at (x, y).

        Parameters
        ----------
        x : int
            X-coordinate of the central pixel.
        y : int
            Y-coordinate of the central pixel.
        size : int, optional
            Size of the grid (must be odd). Default is 3.
        xsize : int, optional
            Number of columns in the grid. Default is size.
        ysize : int, optional
            Number of rows in the grid. Default is size.
        xlim : tuple, optional
            X-axis limits for the plot, in their native units.
        ylim : tuple, optional
            Y-axis limits for the plot, in their native units.
        figsize : tuple, optional
            Size of the figure.
        **kwargs : dict
            Additional keyword arguments passed to `plot_spec_grid`.
        """
        self.get_plotter()
        self.plotter.plot_spec_grid(x, y, size=size, xsize=xsize, ysize=ysize, xlim=xlim, ylim=ylim, figsize=figsize,
                                    **kwargs)

    def plot_fit(self, x, y, ncomp, ax=None, **kwargs):
        """
        Plot a model fit for a spectrum at (x, y).

        Parameters
        ----------
        x : int
            X-coordinate of the pixel.
        y : int
            Y-coordinate of the pixel.
        ncomp : int
            The component number to plot.
        ax : matplotlib.axes.Axes, optional
            Axes object to plot on. If None, a new figure and axes are created.
        **kwargs : dict
            Additional keyword arguments passed to `plot_fit`.
        """
        self.get_plotter()
        if ax is None:
            fig, ax = self.plot_spec(x, y)
        self.plotter.plot_fit(x, y, ax, ncomp, **kwargs)

    def plot_fits_grid(self, x, y, ncomp, size=3, xsize=None, ysize=None, xlim=None, ylim=None,
                       figsize=None, origin='lower', mod_all=True, savename=None, **kwargs):
        """
        Plot a grid of model fits centered at (x, y).

        Parameters
        ----------
        x : int
            X-coordinate of the central pixel.
        y : int
            Y-coordinate of the central pixel.
        ncomp : int
            The component number to plot.
        size : int, optional
            Size of the grid (must be odd). Default is 3.
        xsize : int, optional
            Number of columns in the grid. Default is size.
        ysize : int, optional
            Number of rows in the grid. Default is size.
        xlim : tuple, optional
            X-axis limits for the plot, in their native units.
        ylim : tuple, optional
            Y-axis limits for the plot, in their native units.
        figsize : tuple, optional
            Size of the figure.
        origin : {'lower', 'upper'}, optional
            Origin of the grid. Default is 'lower'.
        mod_all : bool, optional
            Whether to plot all model components. Default is True.
        savename : str, optional
            If provided, save the figure to the given filename.
        **kwargs : dict
            Additional keyword arguments passed to `plot_fits_grid`.
        """
        self.get_plotter()
        self.plotter.plot_fits_grid(x, y, ncomp, size=size, xsize=xsize, ysize=ysize, xlim=xlim, ylim=ylim,
                                    figsize=figsize, origin=origin, mod_all=mod_all, savename=savename, **kwargs)


class UCubePlus(UltraCube):
    """
    A subclass of UltraCube that includes directory management for parameter maps and model fits.
    """
    __module__ = "mufasa.UltraCube"  # Explicitly set the module

    def __init__(self, cubefile, cube=None, fittype=None, paraNameRoot=None, paraDir=None, **kwargs):
        """
        Initialize the UCubePlus object.

        Parameters
        ----------
        cubefile : str
            Path to the .fits cube file.
        cube : SpectralCube, optional
            A spectral cube object. Used if `cubefile` is not provided.
        paraNameRoot : str, optional
            Root name for the parameter map files. If None, the cube file name is used as the basis.
        paraDir : str, optional
            Directory to store the parameter map files. If None, a default directory is created.
        fittype : str, optional
            Keyword for the spectral model to be fitted.
        **kwargs
            Additional keyword arguments passed to the UltraCube initializer.

        Returns
        -------
        None
        """
        super().__init__(cubefile, cube, fittype=fittype, **kwargs)

        self.cubeDir = os.path.dirname(cubefile)

        if paraNameRoot is None:
            # use the cube file name as the basis
            self.paraNameRoot = "{}_paramaps".format(os.path.splitext(os.path.basename(cubefile))[0])
        else:
            self.paraNameRoot = paraNameRoot

        if paraDir is None:
            self.paraDir = "{}/para_maps".format(self.cubeDir)
        else:
            self.paraDir = paraDir

        if not os.path.exists(self.paraDir):
            os.makedirs(self.paraDir)

        self.paraPaths = {}

    def read_model_fit(self, ncomps, read_conv=False, **kwargs):
        """
        Load model fits if they exist; otherwise, perform the fitting.

        .. :noindex:

        Parameters
        ----------
        ncomps : list of int
            List of the number of components to load or fit.
        read_conv : bool, optional
            If True, attempts to read fits for convolved cubes as well. Default is False.
        **kwargs : dict, optional
            Additional keyword arguments passed to the fitting methods.

        Returns
        -------
        None

        Notes
        -----
        - If fits files are not found, the fitting process is triggered, and results are saved to disk.
        - File paths for each component are stored in `self.paraPaths`.

        Raises
        ------
        FileNotFoundError
            If the specified fit files are not found and fitting cannot proceed.
        """
        for nc in ncomps:
            if str(nc) not in self.paraPaths:
                self.paraPaths[str(nc)] = '{}/{}_{}vcomp.fits'.format(self.paraDir, self.paraNameRoot, nc)

            super().load_model_fit(self.paraPaths[str(nc)], ncomp=nc, calc_model=False)

            if 'conv' in self.paraPaths[str(nc)]:
                logger.info(f'Reading convolved cube fits for {nc} component(s)')


    def get_model_fit(self, ncomp, update=True, **kwargs):
        """
        Load the model fits if they exist, or perform fitting if they don't.

        Parameters
        ----------
        ncomp : int or list of int
            Number of components for the model fit. If a list is provided, a fits will be performed for each component.
        update : bool, optional
            Whether to update (i.e., re-fit) the cube even if model fits already exist (default is True).
        **kwargs
            Additional keyword arguments passed to `pyspeckit.Cube.fiteach` if the fitting needs to be updated.

        Returns
        -------
        None
        """
        for nc in ncomp:
            if not str(nc) in self.paraPaths:
                self.paraPaths[str(nc)] = '{}/{}_{}vcomp.fits'.format(self.paraDir, self.paraNameRoot, nc)

        if update:
            # re-fit the cube
            for nc in ncomp:
                if 'conv' in self.paraPaths[str(nc)]:
                    logger.info(f'Fitting convolved cube for {nc} component(s)')
                else:
                    logger.info(f'Fitting cube for {nc} component(s)')
                if 'multicore' not in kwargs:
                    kwargs['multicore'] = self.n_cores
                self.fit_cube(ncomp=[nc], **kwargs)
                gc.collect()
                self.save_fit(self.paraPaths[str(nc)], nc)
                gc.collect()
        else:
            if 'conv' in self.paraPaths[str(nc)]:
                logger.info(f'Loading convolved cube fits for {nc} component(s)')
            else:
                logger.info(f'Loading fits for {nc} component(s)')

        for nc in ncomp:
            path = self.paraPaths[str(nc)]
            self.load_model_fit(path, nc)


#======================================================================================================================#

def fit_cube(cube, pcube, fittype, simpfit=False, **kwargs):
    """
    Fit the spectral cube using the specified fitting type.

    Parameters
    ----------
    cube :
        The cube to be fitted.
    fittype : str
        The type of spectral model to be used for fitting.
    simpfit : bool, optional
        If True, use a simplified fitting method (`cubefit_simp`) without pre-processing (default is False).
    **kwargs
        Additional keyword arguments passed to `pyspeckit.Cube.fiteach`.

    Returns
    -------
    pyspeckit.Cube
        The fitted pyspeckit cube.
    """
    if simpfit:
        # fit the cube with the provided guesses and masks with no pre-processing
        return mvf.cubefit_simp(cube, pcube, fittype=fittype, **kwargs)
    else:
        return mvf.cubefit_gen(cube, pcube, fittype=fittype, **kwargs)



def save_fit(pcube, savename, ncomp, header_note=None):
    """
    Save the fitted parameter cube to a .fits file with the appropriate header.

    Parameters
    ----------
    pcube : pyspeckit.Cube
        The fitted parameter cube to be saved.
    savename : str
        The path where the .fits file will be saved.
    ncomp : int
        The number of components in the model.
    header_note : str, optional
        A single-line comment to include in the FITS header metadata.

    Returns
    -------
    None
    """
    # specifically save ammonia multi-component model with the right fits header
    mvf.save_pcube(pcube, savename, ncomp, header_note=header_note)



def load_model_fit(cube, filename, ncomp, fittype):
    """
    Load the spectral fit results from a .fits file.

    Parameters
    ----------
    cube : SpectralCube
        The spectral cube object to which the fit results will be loaded.
    filename : str
        Path to the .fits file containing the fitted parameters.
    ncomp : int
        Number of components in the model.
    fittype : str
        The keyword that designates the model. Currently available options are 'nh3_multi_v' and 'n2hp_multi_v'.

    Returns
    -------
    pyspeckit.Cube
        The fitted pyspeckit cube with the loaded model.
    """
    #pcube = pyspeckit.Cube(cube=cube)
    pcube = pyspeckit.Cube(cube)

    # register fitter
    if fittype == 'nh3_multi_v':
        linename = 'oneone'
        from .spec_models import ammonia_multiv as ammv
        fitter = ammv.nh3_multi_v_model_generator(n_comp=ncomp, linenames=[linename])

    elif fittype == 'n2hp_multi_v':
        linename = 'onezero'
        from .spec_models import n2hp_multiv as n2hpmv
        fitter = n2hpmv.n2hp_multi_v_model_generator(n_comp=ncomp, linenames=[linename])

    pcube.specfit.Registry.add_fitter(fittype, fitter, fitter.npars)
    pcube.xarr.velocity_convention = 'radio'
    pcube.load_model_fit(filename, npars=fitter.npars, fittype=fittype)
    gc.collect()
    return pcube



def convolve_sky_byfactor(cube, factor, savename=None, **kwargs):
    """
    Convolve the spatial dimensions of a spectral cube by a specified factor.

    This function reduces the spatial resolution of the input cube by convolving
    it with a Gaussian kernel scaled by the given factor. The resulting cube can
    optionally be saved to disk.

    Parameters
    ----------
    cube : SpectralCube
        The input spectral cube to be convolved. Must be an instance of
        `spectral_cube.SpectralCube`.
    factor : int
        The factor by which to reduce the spatial resolution. A factor of 2
        doubles the beam size, effectively halving the spatial resolution.
    savename : str, optional
        The file path to save the convolved cube. If None, the convolved cube
        is not saved. Default is None.
    **kwargs : dict, optional
        Additional keyword arguments passed to the convolution function
        (e.g., to customize the kernel or handle edge cases).

    Returns
    -------
    convolved_cube : SpectralCube
        The convolved spectral cube with reduced spatial resolution.

    Notes
    -----
    - Convolution is performed only on the spatial dimensions of the cube.
    - The kernel size is determined automatically based on the specified `factor`.
    - This function can be computationally intensive for large cubes. Consider
      using memory-efficient techniques if applicable.

    Raises
    ------
    ValueError
        If the input `cube` is not compatible with the convolution operation
        or if the specified `factor` is invalid.
    IOError
        If an error occurs while saving the convolved cube to the specified path.

    """
    return cnvtool.convolve_sky_byfactor(cube, factor, savename, **kwargs)

#======================================================================================================================#
# UltraCube based methods

def calc_rss(ucube, compID, usemask=True, mask=None, return_size=True, update_cube=False, planemask=None, expand=20):
    """
    Calculate the residual sum of squares (RSS) for a spectral cube model fit.

    The RSS is computed as the sum of squared differences between the observed data
    and the model values for each voxel in the spectral cube. The calculation can be
    restricted using masks and expanded spectral regions.

    Parameters
    ----------
    ucube : UltraCube
        An instance of the `UltraCube` class containing the spectral data and fitted models.
    compID : int or str
        The component ID or number of components in the model to evaluate. If `compID` is 0,
        the model is assumed to be a flat baseline (`y = 0`).
    usemask : bool, optional
        Whether to apply a mask during the RSS calculation. If True, regions where the
        model is zero or invalid are excluded. Default is True.
    mask : numpy.ndarray, optional
        A 3D boolean array specifying which voxels to include in the calculation.
        If None and `usemask` is True, a default mask derived from the model is used.
    return_size : bool, optional
        Whether to return the effective sample size (number of valid voxels) for each spatial pixel.
        Default is True.
    update_cube : bool, optional
        Whether to update the `UltraCube` instance with the calculated RSS values. Default is False.
    planemask : numpy.ndarray, optional
        A 2D spatial boolean array specifying which spatial pixels to calculate RSS for.
        If None, all spatial pixels are considered. Default is None.
    expand : int, optional
        Number of spectral channels to expand the mask region for RSS calculation. Default is 20.

    Returns
    -------
    rss : numpy.ndarray
        A 2D array containing the RSS values for each spatial pixel.
    nsamp : numpy.ndarray, optional
        A 2D array containing the effective sample size (number of valid spectral samples)
        for each spatial pixel. Returned only if `return_size` is True.
    final_mask : numpy.ndarray, optional
        The 3D boolean mask used in the RSS calculation. Returned only if explicitly required
        by the caller.

    Notes
    -----
    The RSS is calculated as:

        RSS = Σ (data - model)²

    where the sum is taken over spectral channels for each spatial pixel.

    If `expand > 0`, the mask region is extended by the specified number of spectral channels.

    Raises
    ------
    ValueError
        If the dimensions of the data cube and model cube do not match, or if an invalid mask is provided.
    """
    if isinstance(compID, int):
        compID = str(compID)

    cube = ucube.cube

    if compID == '0':
        # the zero component model is just a y = 0 baseline
        modcube = np.zeros(cube.shape)
    else:
        modcube = ucube.pcubes[compID].get_modelcube(update=False, multicore=ucube.n_cores)

    gc.collect()
    return get_rss(cube, modcube, expand=expand, usemask=usemask, mask=mask, return_size=return_size,
                   planemask=planemask)


def calc_chisq(ucube, compID, reduced=False, usemask=False, mask=None, expand=20):
    """
    Calculate the chi-squared (χ²) or reduced chi-squared value for a spectral cube model fit.

    This function computes the goodness-of-fit by comparing the data in the spectral cube
    to the model values. Optionally, the calculation can include masking and spectral
    region expansion.

    Parameters
    ----------
    ucube : UltraCube
        An instance of the `UltraCube` class containing the spectral data and fitted models.
    compID : int or str
        The component ID or number of components in the model to evaluate. If `compID` is 0,
        the model is assumed to be a flat baseline (`y = 0`).
    reduced : bool, optional
        Whether to compute the reduced chi-squared value by normalizing with the degrees of freedom.
        If False, computes the standard chi-squared value. Default is False.
    usemask : bool, optional
        Whether to apply a mask to exclude invalid or unwanted regions from the calculation. Default is False.
    mask : numpy.ndarray, optional
        A 3D boolean array specifying which voxels to include in the chi-squared calculation.
        If None and `usemask` is True, a default mask derived from the model is used.
    expand : int, optional
        Number of spectral channels to expand the mask region. Default is 20.

    Returns
    -------
    chisq : numpy.ndarray
        A 2D array representing the chi-squared (or reduced chi-squared) values
        for each spatial pixel in the cube.

    Notes
    -----
    The chi-squared statistic is calculated as:

        χ² = Σ [(data - model)² / rms²]

    where the sum is taken over spectral channels for each spatial pixel.

    If `reduced=True`, the reduced chi-squared is computed by dividing χ² by
    the degrees of freedom, which is the number of samples minus the number of model parameters.

    Raises
    ------
    ValueError
        If the dimensions of the data cube and model cube do not match.
    """
    if isinstance(compID, int):
        compID = str(compID)

    cube = ucube.cube

    if compID == '0':
        # the zero component model is just a y = 0 baseline
        modcube = np.zeros(cube.shape)
    else:
        modcube = ucube.pcubes[compID].get_modelcube(multicore=ucube.n_cores)

    gc.collect()
    return get_chisq(cube, modcube, expand=expand, reduced=reduced, usemask=usemask, mask=mask)


def calc_AICc(ucube, compID, mask, planemask=None, return_NSamp=True, expand=20):
    """
    Calculate the corrected Akaike Information Criterion (AICc) for a spectral cube model.

    This function computes the AICc for a given model component by evaluating
    the residual sum of squares (RSS) and the effective sample size. The AICc
    values help compare the goodness-of-fit for models with varying complexities.

    Parameters
    ----------
    ucube : UltraCube
        An instance of the `UltraCube` class containing the spectral data, fitted models,
        and associated parameters.
    compID : int or str
        The component ID or number of components in the model. If an integer,
        it is used to calculate the number of parameters as `n_parameters = compID * 4`.
    mask : numpy.ndarray
        A 3D boolean array specifying which voxels (volume pixels) to include in the AICc calculation.
    planemask : numpy.ndarray, optional
        A 2D spatial boolean array specifying which pixels to calculate AICc for. If provided,
        calculations are restricted to these pixels. Default is None.
    return_NSamp : bool, optional
        If True, return the effective sample size along with the AICc values. Default is True.
    expand : int, optional
        Number of spectral channels to expand the region where the RSS is calculated.
        Default is 20.

    Returns
    -------
    AICc_map : numpy.ndarray
        A 2D array containing the computed AICc values for the specified regions.
    NSamp_map : numpy.ndarray, optional
        A 2D array containing the effective sample size for each spatial pixel.
        Only returned if `return_NSamp` is True.

    Notes
    -----
    The corrected Akaike Information Criterion (AICc) is a modification of the AIC
    that accounts for finite sample sizes. It is given by:

        AICc = AIC + (2k(k + 1)) / (N - k - 1)

    where `k` is the number of parameters and `N` is the sample size.

    Raises
    ------
    ValueError
        If the input `ucube` does not contain valid models or residuals for the specified `compID`.
    """
    if isinstance(compID, int):
        p = compID * 4
        compID = str(compID)
    elif isinstance(compID, str):
        p = int(compID) * 4

    if compID == '0':
        # the zero component model is just a y = 0 baseline
        modcube = np.zeros(cube.shape)
    else:
        modcube = ucube.pcubes[compID].get_modelcube(update=False, multicore=ucube.n_cores)

    # get the rss value and sample size
    rss_map, NSamp_map = get_rss(ucube.cube, modcube, expand=expand, usemask=True, mask=mask,
                                 return_size=True, return_mask=False, planemask=planemask)
    AICc_map = aic.AICc(rss=rss_map, p=p, N=NSamp_map)

    if return_NSamp:
        return AICc_map, NSamp_map
    else:
        return AICc_map


def calc_AICc_likelihood(ucube, ncomp_A, ncomp_B, ucube_B=None, multicore=True, expand=0, planemask=None):
    """
    Calculate the relative likelihood of two models based on their AICc values.

    This function computes the logarithmic relative likelihood of model A
    compared to model B, given their corrected Akaike Information Criterion (AICc)
    values. It can optionally use a second `UltraCube` instance for comparisons.

    Parameters
    ----------
    ucube : UltraCube
        An instance of the `UltraCube` class containing the spectral data,
        fitted models, and associated parameters for model A.
    ncomp_A : int
        Number of components in model A.
    ncomp_B : int
        Number of components in model B.
    ucube_B : UltraCube, optional
        An optional second `UltraCube` instance for model B. If provided,
        AICc values for both models are calculated independently, and a
        common mask is used. Default is None.
    multicore : bool, optional
        Whether to enable parallel processing. Default is True.
    expand : int, optional
        Number of spectral channels to expand the region where AICc values
        are calculated. Default is 0.
    planemask : numpy.ndarray, optional
        A 2D spatial boolean array specifying which pixels to calculate the
        relative likelihood for. If None, all pixels are considered. Default is None.

    Returns
    -------
    lnk : numpy.ndarray
        A 2D array containing the logarithmic relative likelihood of model A
        compared to model B.

    Notes
    -----
    - The likelihood is derived from the difference in AICc values:

        .. math::

            \ln(\mathcal{L}_A / \mathcal{L}_B) = (AICc_B - AICc_A) / 2

      where :math:`\mathcal{L}_A` and :math:`\mathcal{L}_B` are the likelihoods of models A and B.
    - If `ucube_B` is provided, the models are compared over their common mask, and the AICc values
      are recalculated fresh.
    - This function handles mismatches in sample size (`NSamp`) by resetting and updating model masks.
    - Currently, the expand argument is only used if ucube_B is provided

    Raises
    ------
    ValueError
        If the required AICc values for the models cannot be calculated due to missing data or invalid masks.

    """
    if not ucube_B is None:
        # if a second UCube is provide for model comparison, use their common mask and calculate AICc values
        # without storing/updating them in the UCubes
        # reset model masks first
        ucube.reset_model_mask(ncomps=[ncomp_A], multicore=multicore)
        ucube_B.reset_model_mask(ncomps=[ncomp_B], multicore=multicore)

        mask = np.logical_or(ucube.master_model_mask, ucube_B.master_model_mask)
        AICc_A = calc_AICc(ucube, compID=ncomp_A, mask=mask, planemask=planemask, return_NSamp=False, expand=expand)
        AICc_B = calc_AICc(ucube_B, compID=ncomp_B, mask=mask, planemask=planemask, return_NSamp=False, expand=expand)

        return aic.likelihood(AICc_A, AICc_B)

    if not str(ncomp_A) in ucube.NSamp_maps:
        ucube.get_AICc(ncomp_A)

    if not str(ncomp_B) in ucube.NSamp_maps:
        ucube.get_AICc(ncomp_B)

    NSamp_mapA = ucube.NSamp_maps[str(ncomp_A)]
    NSamp_mapB = ucube.NSamp_maps[str(ncomp_B)]

    if not np.array_equal(NSamp_mapA, NSamp_mapB, equal_nan=True):
        logger.warning("Number of samples do not match. Recalculating AICc values")
        #reset the master component mask first
        ucube.reset_model_mask(ncomps=[ncomp_A, ncomp_B], multicore=multicore)

        if planemask is None:
            pmask = NSamp_mapA != NSamp_mapB
        else:
            pmask = np.logical_and(planemask, NSamp_mapA != NSamp_mapB)
        ucube.get_AICc(ncomp_A, update=True, planemask=pmask)
        ucube.get_AICc(ncomp_B, update=True, planemask=pmask)

    gc.collect()

    if planemask is None:
        lnk = aic.likelihood(ucube.AICc_maps[str(ncomp_A)], ucube.AICc_maps[str(ncomp_B)])
    else:
        lnk = aic.likelihood(ucube.AICc_maps[str(ncomp_A)][planemask], ucube.AICc_maps[str(ncomp_B)][planemask])
    return lnk

def get_all_lnk_maps(ucube, ncomp_max=2, rest_model_mask=True, multicore=True):
    """
    Compute log-likelihood ratio maps for model comparisons up to a specified number of components.

    This function calculates log-likelihood ratio (lnk) maps for comparing spectral
    models with different numbers of components, based on the Akaike Information
    Criterion corrected for finite sample sizes (AICc).

    Parameters
    ----------
    ucube : UltraCube
        An instance of the `UltraCube` class containing the spectral data and fitted models.
    ncomp_max : int, optional
        The maximum number of components to include in the model comparison.
        Default is 2.
    rest_model_mask : bool, optional
        If True, resets and updates the master model mask in `ucube` for components
        being compared. Default is True.
    multicore : bool, optional
        Whether to enable parallel processing for calculations. Default is True.

    Returns
    -------
    lnk_maps : tuple of numpy.ndarray
        Log-likelihood ratio maps for the model comparisons. The returned maps include:
        - `lnk10`: Comparison between 1-component and 0-component models.
        - `lnk20` (if `ncomp_max >= 2`): Comparison between 2-component and 0-component models.
        - `lnk21` (if `ncomp_max >= 2`): Comparison between 2-component and 1-component models.

    Notes
    -----
    - Log-likelihood ratios are calculated using AICc values for each model as:

        .. math::

            \ln(\mathcal{L}_A / \mathcal{L}_B) = (AICc_B - AICc_A) / 2

      where :math:`\mathcal{L}_A` and :math:`\mathcal{L}_B` are the likelihoods of
      models A and B, respectively.
    - If `ncomp_max` is greater than 2, the function does not compute higher-order
      comparisons and simply returns the log-likelihood maps for up to 2 components.

    Raises
    ------
    ValueError
        If `ncomp_max` is less than 1 or if model data for the required number of
        components is missing in `ucube`.
    """
    if rest_model_mask:
        ucube.reset_model_mask(ncomps=[2, 1], multicore=multicore)

    if ncomp_max <=1:
        lnk10 = ucube.get_AICc_likelihood(1, 0)
        return lnk10

    if ncomp_max <= 2:
        lnk21 = ucube.get_AICc_likelihood(2, 1)
        lnk20 = ucube.get_AICc_likelihood(2, 0)
        lnk10 = ucube.get_AICc_likelihood(1, 0)
        return lnk10, lnk20, lnk21

    else:
        pass


def get_best_2c_parcube(ucube, multicore=True, lnk21_thres=5, lnk20_thres=5, lnk10_thres=5, return_lnks=True, include_1c=True):
    """
    Select the best 2-component parameter cube based on AICc likelihood thresholds.

    This function identifies regions in the parameter cube where the 2-component model
    is justified over simpler models (0- or 1-component) based on log-likelihood ratio
    thresholds calculated from the AICc. It can also combine results with a 1-component
    model for regions where 2-component fits are not justified.

    Parameters
    ----------
    ucube : UltraCube
        An instance of the `UltraCube` class containing spectral data and model fits.
    multicore : bool, optional
        Whether to enable parallel processing for calculations. Default is True.
    lnk21_thres : float, optional
        Threshold for the log-likelihood ratio between 2-component and 1-component models.
        Default is 5.
    lnk20_thres : float, optional
        Threshold for the log-likelihood ratio between 2-component and 0-component models.
        Default is 5.
    lnk10_thres : float, optional
        Threshold for the log-likelihood ratio between 1-component and 0-component models.
        Default is 5.
    return_lnks : bool, optional
        If True, return the log-likelihood ratio maps (lnk10, lnk20, lnk21) along with
        the parameter and error cubes. Default is True.
    include_1c : bool, optional
        If True, include 1-component model results in regions where the 2-component model
        is not justified. Default is True.

    Returns
    -------
    tuple
        If `return_lnks` is True:
            - parcube : numpy.ndarray
                A 3D array representing the best-fit parameter cube for the justified
                model in each spatial region.
            - errcube : numpy.ndarray
                A 3D array representing the associated errors for the parameters.
            - lnk10 : numpy.ndarray
                Log-likelihood ratio map comparing 1-component and 0-component models.
            - lnk20 : numpy.ndarray
                Log-likelihood ratio map comparing 2-component and 0-component models.
            - lnk21 : numpy.ndarray
                Log-likelihood ratio map comparing 2-component and 1-component models.
        If `return_lnks` is False:
            - parcube : numpy.ndarray
            - errcube : numpy.ndarray

    Notes
    -----
    - The function determines the best model for each spatial region by comparing
      log-likelihood ratios with the specified thresholds.
    - Regions where the 2-component model is not justified (based on `lnk21_thres`
      and `lnk20_thres`) can revert to the 1-component model if `include_1c` is True.
    - Any region failing to meet the `lnk10_thres` threshold for the 1-component model
      is set to NaN in the parameter and error cubes.

    Raises
    ------
    ValueError
        If model fits for 1-component or 2-component models are missing in `ucube`.

    """
    lnk10, lnk20, lnk21 = get_all_lnk_maps(ucube, ncomp_max=2, multicore=multicore)

    parcube = copy(ucube.pcubes['2'].parcube)
    errcube = copy(ucube.pcubes['2'].errcube)

    mask = np.logical_and(lnk21 > lnk21_thres, lnk20 > lnk20_thres)

    if include_1c:
        parcube[:4, ~mask] = copy(ucube.pcubes['1'].parcube[:4, ~mask])
        errcube[:4, ~mask] = copy(ucube.pcubes['1'].errcube[:4, ~mask])
        parcube[4:8, ~mask] = np.nan
        errcube[4:8, ~mask] = np.nan

    else:
        parcube[:, ~mask] = np.nan
        errcube[:, ~mask] = np.nan

    mask = lnk10 > lnk10_thres
    parcube[:, ~mask] = np.nan
    errcube[:, ~mask] = np.nan

    if return_lnks:
        # set lnk pixels with no fit to nan
        mask = ucube.has_fit(ncomp=1)
        lnk10[~mask] = np.nan
        mask = ucube.has_fit(ncomp=2)
        lnk20[~mask] = np.nan
        lnk21[~mask] = np.nan

        return parcube, errcube, lnk10, lnk20, lnk21
    else:
        return parcube, errcube

#======================================================================================================================#
# statistics tools

def get_rss(cube, model, expand=20, usemask=True, mask=None, return_size=True, return_mask=False, include_nosamp=True, planemask=None):
    """
    Calculate the residual sum of squares (RSS) for a spectral cube model fit.

    Parameters
    ----------
    cube : SpectralCube
        The spectral data cube to analyze.
    model : numpy.ndarray
        The model array to compare against the data cube.
    expand : int, optional
        Number of channels to expand the region where the RSS is calculated along the spectral dimension. Default is 20.
    usemask : bool, optional
        Whether to apply a mask during RSS calculation. Default is True.
    mask : numpy.ndarray, optional
        A boolean array specifying which elements to include in the calculation. If None, a default mask based on the model is used.
    return_size : bool, optional
        If True, return the size of the valid sample region used in the RSS calculation. Default is True.
    return_mask : bool, optional
        If True, return the final mask used during RSS computation. Default is False.
    include_nosamp : bool, optional
        Whether to include spectral regions with no sample data by filling gaps with a default mask. Default is True.
    planemask : numpy.ndarray, optional
        A 2D mask specifying where to calculate RSS for optimized computation. Default is None.

    Returns
    -------
    tuple
        The calculated RSS values. If `return_size` or `return_mask` is True, additional elements include:
        - Sample size per pixel (if `return_size` is True).
        - Final mask used (if `return_mask` is True).
    """
    if usemask:
        if mask is None:
            # may want to change this for future models that includes absorptions
            mask = model > 0
    else:
        mask = ~np.isnan(model)

    if include_nosamp:
        # if there no mask in a given pixel, fill it in with combined spectral mask
        nsamp_map = np.nansum(mask, axis=0)
        mm = nsamp_map <= 0
        try:
            max_y, max_x = np.where(nsamp_map == np.nanmax(nsamp_map))
            spec_mask_fill = copy(mask[:, max_y[0], max_x[0]])
        except:
            spec_mask_fill = np.any(mask, axis=(1, 2))

        # Handle dask compatibility by computing the mask if necessary
        if isinstance(mask, da.Array):
            mask = mask.compute()

        # Apply the mask update with numpy-compatible indexing
        mask[:, mm] = spec_mask_fill[:, np.newaxis]

        # Convert back to dask if needed
        if isinstance(cube._data, da.Array):
            mask = da.from_array(mask, chunks=cube._data.chunksize)

    # assume flat-baseline model even if no model exists
    model[np.isnan(model)] = 0

    # creating mask over region where the model is non-zero,
    # plus a buffer of size set by the expand keyword.
    if expand > 0:
        mask = expand_mask(mask, expand)
    mask = mask.astype(float)

    if planemask is None:
        residual = get_residual(cube, model)
        residual = da.from_array(residual) if isinstance(cube._data, da.Array) else residual
    else:
        residual = get_residual(cube, model, planemask=planemask)
        mask_temp = mask
        mask = mask[:, planemask]

    # note: using nan-sum may walk over some potential bad pixel cases
    rss = da.nansum((residual * mask) ** 2, axis=0) if isinstance(residual, da.Array) else np.nansum(
        (residual * mask) ** 2, axis=0)
    rss[rss == 0] = np.nan

    returns = (rss.compute() if isinstance(rss, da.Array) else rss,)

    if return_size:
        nsamp = da.nansum(mask, axis=0) if isinstance(mask, da.Array) else np.nansum(mask, axis=0)
        nsamp[np.isnan(rss)] = np.nan
        returns += (nsamp.compute() if isinstance(nsamp, da.Array) else nsamp,)
    if return_mask:
        returns += (mask.compute() if isinstance(mask, da.Array) else mask,)

    return returns


def get_chisq(cube, model, expand=20, reduced=True, usemask=True, mask=None):
    """
    Calculate the chi-squared or reduced chi-squared value for a spectral cube.

    This function computes the chi-squared goodness-of-fit statistic by comparing
    the observed data in the cube with a provided model. Optionally, the calculation
    can be restricted to masked regions and expanded spectral regions.

    Parameters
    ----------
    cube : SpectralCube
        The observed spectral cube containing the data to compare against the model.
    model : numpy.ndarray
        A 3D array representing the model cube, which must have the same shape as the input cube.
    expand : int, optional
        Number of spectral channels to expand the mask region. Default is 20.
    reduced : bool, optional
        If True, compute the reduced chi-squared value by normalizing with the degrees of freedom.
        If False, compute the standard chi-squared value. Default is True.
    usemask : bool, optional
        If True, apply a mask to exclude invalid regions from the calculation. If no mask is provided,
        regions where the model is zero are excluded. Default is True.
    mask : numpy.ndarray, optional
        A 3D boolean array specifying regions to include in the calculation. If None,
        a mask is automatically derived from the model.

    Returns
    -------
    numpy.ndarray
        A 2D array representing the chi-squared (or reduced chi-squared) value
        at each spatial pixel in the cube.

    Notes
    -----
    - The `expand` parameter allows users to extend the region of interest by a
      specified number of spectral channels around the model.
    - The reduced chi-squared value is normalized by the number of degrees of freedom.
    - The residuals between the cube and model are weighted using the mask.

    Raises
    ------
    ValueError
        If the cube and model dimensions do not match.

    Examples
    --------
    >>> from spectral_cube import SpectralCube
    >>> import numpy as np
    >>> cube = SpectralCube.read("example_cube.fits")
    >>> model = np.random.random(cube.shape)
    >>> chisq = get_chisq(cube, model, expand=10, reduced=True)
    >>> print(chisq)
    """
    if usemask:
        if mask is None:
            mask = model > 0
    else:
        mask = ~np.isnan(model)

    residual = get_residual(cube, model)

    # creating mask over region where the model is non-zero,
    # plus a buffer of size set by the expand keyword.

    if expand > 0:
        mask = expand_mask(mask, expand)
    mask = mask.astype(float)

    # note: using nan-sum may walk over some potential bad pixel cases
    chisq = np.nansum((residual * mask) ** 2, axis=0)

    if reduced:
        # assuming n_size >> n_parameters
        reduction = np.nansum(mask, axis=0) # avoid division by zero
        reduction[reduction == 0] = np.nan
        chisq /= reduction

    rms = get_rms(residual)
    chisq /= rms ** 2

    gc.collect()

    if reduced:
        # return the reduce chi-squares values
        return chisq
    else:
        # return the ch-squared values and the number of data points used
        return chisq, np.nansum(mask, axis=0)


def get_masked_moment(cube, model, order=0, expand=10, mask=None):
    """
    Calculate a masked moment of a spectral cube.

    This function generates a moment map (e.g., integrated intensity, centroid)
    for the input spectral cube, applying a mask derived from the provided model
    and optionally expanding the mask region.

    Parameters
    ----------
    cube : SpectralCube
        The spectral cube containing the observed data, from which the moment
        is calculated.
    model : numpy.ndarray
        A 3D array representing the model cube, used to derive the mask. The
        dimensions must match the cube.
    order : int, optional
        The order of the moment to compute:
        - `0` for integrated intensity (default),
        - `1` for centroid,
        - `2` for line width.
    expand : int, optional
        Number of spectral channels to expand the mask around the model.
        Default is 10.
    mask : numpy.ndarray, optional
        A 3D boolean array specifying which elements to include in the moment
        calculation. If provided, it is combined with the mask derived from
        the model. Default is None.

    Returns
    -------
    astropy.units.Quantity
        The computed moment map, with the unit determined by the input cube
        and the moment order.

    Notes
    -----
    - The mask is generated by identifying non-zero elements in the model and
      optionally expanding this region by the `expand` parameter.
    - Pixels with low signal-to-noise ratios are excluded from the moment map
      based on the mask.
    - The function uses the spectral axis of the cube for moment calculations.

    Raises
    ------
    ValueError
        If the cube and model dimensions do not match.

    Examples
    --------
    >>> from spectral_cube import SpectralCube
    >>> cube = SpectralCube.read("example_cube.fits")
    >>> model = np.random.random(cube.shape)
    >>> moment_map = get_masked_moment(cube, model, order=0, expand=5)
    >>> moment_map.write("masked_moment.fits")
    """
    if mask is None:
        mask = model > 0
    else:
        mask = np.logical_and(mask, np.isfinite(model))

    # get mask over where signal is stronger than the median
    peak_T = np.nanmax(model, axis=0)
    med_peak_T = np.nanmedian(peak_T)
    mask_highT_2d = peak_T > med_peak_T

    mask_lowT = mask.copy()
    mask_lowT[:, mask_highT_2d] = False

    # get all the spectral channels greater than 10% of the median peak
    specmask = model > med_peak_T*0.1
    specmask = np.any(specmask, axis=(1,2))

    # adopte those spectral channles for low signal regions
    mask_lowT[specmask, :] = True
    mask[:, ~mask_highT_2d] = mask_lowT[:, ~mask_highT_2d]

    # creating mask over region where the model is non-zero,
    # plus a buffer of size set by the expand keyword.

    if expand > 0:
        mask = expand_mask(mask, expand)
    mask = mask.astype(float)

    maskcube = cube.with_mask(mask.astype(bool))
    maskcube = maskcube.with_spectral_unit(u.km / u.s, velocity_convention='radio')
    mom = maskcube.moment(order=order)
    return mom


def expand_mask(mask, expand):
    """
    Expand a 3D mask along the spectral axis by a specified buffer size.

    This function applies a binary dilation to the input mask, increasing its coverage
    along the spectral axis by the specified `expand` value. The expansion is applied
    independently for each spatial pixel.

    Parameters
    ----------
    mask : numpy.ndarray
        A 3D boolean array representing the original mask. The first dimension corresponds
        to the spectral axis, while the remaining dimensions represent spatial axes.
    expand : int
        The number of spectral channels to extend the mask along the spectral axis.
        Must be a non-negative integer.

    Returns
    -------
    expanded_mask : numpy.ndarray
        A 3D boolean array of the same shape as the input mask, with the spectral
        regions expanded by the specified buffer size.

    Notes
    -----
    The binary dilation process enlarges the `True` regions in the mask along the
    spectral axis. The dilation does not affect spatial axes, but it creates a buffer
    around the original mask in the spectral dimension.

    Raises
    ------
    ValueError
        If `expand` is not a non-negative integer.
    TypeError
        If the input `mask` is not a boolean numpy array.

    """
    selem = np.ones(expand,dtype=bool)
    selem.shape += (1,1,)
    mask = nd.binary_dilation(mask, selem)
    return mask


def get_rms(residual):
    """
    Compute a robust estimate of the root mean square (RMS) from the fit residuals.

    This function calculates the RMS of the residuals using a robust method based on the
    median absolute deviation (MAD). It is less sensitive to outliers compared to a
    standard RMS calculation.

    Parameters
    ----------
    residual : numpy.ndarray
        A 3D array representing the residuals between the observed data and the fitted model.
        The first dimension corresponds to the spectral axis, while the remaining dimensions
        represent the spatial axes.

    Returns
    -------
    rms : numpy.ndarray
        A 2D array representing the RMS for each spatial pixel in the residual cube.

    Notes
    -----
    The RMS is calculated using the following robust formula:

    RMS = 1.4826 × median(abs(x - median(x))) / sqrt(2)

    Here, `x` represents the residuals for each spectral channel at a spatial pixel. The
    subtraction and absolute value are applied element-wise along the spectral axis. The robust
    method reduces sensitivity to noise spikes or outliers in the residuals.

    Raises
    ------
    ValueError
        If the input `residual` is not a 3D array or if it contains invalid (e.g., NaN) values.

    """
    diff = residual - np.roll(residual, 2, axis=0)
    rms = 1.4826 * np.nanmedian(np.abs(diff), axis=0) / 2**0.5
    gc.collect()
    return rms


def get_residual(cube, model, planemask=None):
    """
    Calculate the residual between the data cube and the model cube.

    The residual is calculated as the difference between the spectral data
    in the cube and the corresponding model values. Optionally, a 2D
    spatial mask (`planemask`) can be applied to restrict the calculation
    to specific spatial regions.

    Parameters
    ----------
    cube : SpectralCube
        A spectral cube object containing the observed data. It can be a
        dask-enabled cube or a standard numpy-based cube.
    model : numpy.ndarray
        A 3D array representing the model cube, where the dimensions match
        those of the data cube (spectral axis as the first dimension).
    planemask : numpy.ndarray, optional
        A 2D boolean array specifying the spatial pixels for which residuals
        should be calculated. If provided, only these regions will be used
        in the calculation. Default is None.

    Returns
    -------
    numpy.ndarray or dask.array.Array
        The residual array, calculated as the difference between the cube's
        data and the model values. The returned type matches the cube's data
        type (e.g., dask array if the cube uses dask, or numpy array otherwise).

    Notes
    -----
    - If `planemask` is provided, the residuals are calculated only for the
      specified spatial pixels, which can reduce computation time.
    - Handles memory-efficient computation when dask is enabled.
    """
    # Get the cube data as a dask array or numpy array
    data = cube.filled_data[:].value  # dask array if dask is enabled, numpy array otherwise

    # Calculate residual with or without a planemask
    if planemask is None:
        residual = data - model
    else:
        # If dask, apply the mask in a memory-efficient way
        if isinstance(data, da.Array):
            planemask_expanded = da.from_array(planemask, chunks=data.chunksize[1:])
            residual = data[:, planemask_expanded] - model[:, planemask]
        else:
            # Non-dask (numpy array), use direct masking
            residual = data[:, planemask] - model[:, planemask]

    # If residual is a dask array, compute only if needed (e.g., for direct use in numpy context)
    if isinstance(residual, da.Array):
        residual = residual.compute()

    # Run garbage collection for memory management
    gc.collect()

    return residual


def get_Tpeak(model):
    """
    Calculate the peak value of a model cube at each spatial pixel.

    Parameters
    ----------
    model : numpy.ndarray
        The input model cube with spectral data. The first dimension is assumed
        to represent the spectral axis, and subsequent dimensions represent spatial pixels.

    Returns
    -------
    numpy.ndarray
        A 2D array where each element corresponds to the peak value of the model
        cube along the spectral axis for each spatial pixel.
    """
    return np.nanmax(model, axis=0)


def is_K(data_unit):
    """
    Check if a given unit is equivalent to Kelvin (K).

    Parameters
    ----------
    data_unit : astropy.units.Unit or str
        The unit to be checked. It can be an `astropy.units.Unit` instance or a string
        representation of the unit.

    Returns
    -------
    bool
        True if the provided unit is equivalent to Kelvin (K), otherwise False.

    Examples
    --------
    >>> from astropy import units as u
    >>> is_K(u.K)
    True
    >>> is_K('K')
    True
    >>> is_K(u.Jy)
    False
    """
    return data_unit == 'K' or data_unit == u.K


def to_K(cube):
    """
    Convert the unit of a spectral cube to Kelvin (K).

    This function attempts to convert the unit of a `spectral_cube.SpectralCube` object
    to brightness temperature (K) using the Rayleigh-Jeans approximation. If the cube's
    current unit is not convertible, it will handle the error by either warning the user
    or raising a `UnitConversionError`.

    Parameters
    ----------
    cube : spectral_cube.SpectralCube
        A `SpectralCube` object whose unit needs to be converted to Kelvin (K).

    Returns
    -------
    spectral_cube.SpectralCube
        A new `SpectralCube` object with its unit set to Kelvin (K).

    Raises
    ------
    UnitConversionError
        If the cube's unit cannot be converted to Kelvin (K) and it is not possible to
        forcefully assign a unit.

    Notes
    -----
    If the cube does not have an assigned unit, a unit of Kelvin (K) will be assumed and
    a warning will be issued.

    Examples
    --------
    >>> from spectral_cube import SpectralCube
    >>> import astropy.units as u
    >>> cube = SpectralCube.read('example_cube.fits')
    >>> cube_in_K = to_K(cube)
    """
    # Decorator to handle UnitConversionError
    def handle_unit_conversion_error(func):
        @wraps(func)
        def wrapper(cube, *args, **kwargs):
            try:
                return func(cube, *args, **kwargs)
            except UnitConversionError:
                if hasattr(cube, 'unit'):
                    if cube.unit is None or cube.unit == '':
                        logger.warning("The cube does not have a unit. A unit of K will be assumed.")
                    else:
                        raise UnitConversionError(f"Cube's unit ({cube.unit}) is not convertible to K")
                else:
                    logger.warning("The cube does not have a unit. A unit of K will be assumed.")
                cube._unit = u.K
                return cube

        return wrapper

    # Decorator to handle ValueError and retry the conversion
    def handle_value_error(func):
        @wraps(func)
        def wrapper(cube, *args, **kwargs):
            try:
                return func(cube, *args, **kwargs)
            except ValueError:
                # Allow the entire cube to be loaded temporarily if the cube is too large
                logger.warning("ValueError encountered, temporarily allowing huge operations.")
                cube.allow_huge_operations = True
                result = func(cube, *args, **kwargs)
                gc.collect()  # Garbage collection to free up memory
                cube.allow_huge_operations = False
                return result

        return wrapper

    # Function to convert the cube to units of K
    @handle_value_error
    @handle_unit_conversion_error
    def convert_to_kelvin(cube):
        # Only convert if the cube does not already have a unit of K
        if cube.unit != u.K:
            cube = cube.to(u.K)
        return cube

    return convert_to_kelvin(cube)





__author__ = 'mcychen'

import numpy as np
from astropy import units as u
from concurrent.futures import ThreadPoolExecutor

from ..exceptions import FitTypeError
from .. import moment_guess as momgue

class MetaModel(object):
    """
    A class to store spectral model-specific information relevant to spectral modeling tasks, such as fitting.

    Attributes
    ----------
    fittype : str
        The identifying name for the spectral model (i.e., the type of model to fit).
    ncomp : int
        The number of components within the model.
    adoptive_snr_masking : bool
        Whether to use adaptive SNR masking during fitting (default is True).
    window_hwidth : int
        The default half-width of the window for calculating moments.
    central_win_hwidth : float or None
        The half-width of the intermediate window for filtering out satellite hyperfine lines. Default is None.
    Texmin : float
        Lower excitation temperature limit in Kelvin. Default is 3.0 K.
    Texmax : float
        Upper excitation temperature limit in Kelvin. Default is 40 K.
    sigmin : float
        Lower velocity dispersion limit in km/s. Default is 0.07 km/s.
    sigmax : float
        Upper velocity dispersion limit in km/s. Default is 2.5 km/s.
    taumin : float
        Lower optical depth limit. Default is 0.1.
    taumax : float
        Upper optical depth limit. Default is 30.0.
    eps : float
        A small perturbation that can be used in initial guesses. Default is 0.001.
    main_hf_moments : function
        Function for calculating moments for the main hyperfine component.
    linetype : str
        The type of spectral line (e.g., 'nh3', 'n2hp'). Set based on `fittype`.
    linenames : list of str
        Names of the spectral lines being modeled. Set based on `fittype`.
    linename : str
        Name of the main spectral line being modeled. Set based on `fittype`.
    fitter : function
        The model generator function for fitting spectral components. Set based on `fittype`.
    freq_dict : dict
        Dictionary containing rest frequencies for the spectral lines. Set based on `fittype`.
    rest_value : astropy.units.Quantity
        Rest frequency of the spectral line in Hz. Set based on `fittype`.
    voffs : list of float
        Velocity offsets for the hyperfine lines in km/s. Set based on `fittype`.
    tau_wts : list of float
        Weights for the optical depth of hyperfine lines. Set based on `fittype`.
    voff_at_peak : float
        Velocity offset at the brightest hyperfine line in km/s. Set based on `fittype`.
    model_func : function
        The specific spectral model function used for fitting. Set based on `fittype`.
    """
    def __init__(self, fittype, ncomp):
        """
        Initialize a MetaModel object.

        Parameters
        ----------
        fittype : str
            The identifying name for the spectral model (e.g., 'nh3_multi_v', 'n2hp_multi_v').
        ncomp : int
            The number of components within the model to be fitted.

        Raises
        ------
        FitTypeError
            If the provided `fittype` is not recognized.
        """
        self.fittype = fittype
        self.adoptive_snr_masking = True

        # the default windows for calculating moments
        # final window for evaluating moments (center differs from pixel to pixel)
        self.window_hwidth = 5
        # intermediate window for filtering out satellite hyperfine lines. The window is fixed and the same
        # throughout all pixels
        self.central_win_hwidth = None

        # set default parameter limits
        # (primarily based on NH3 (1,1) in the galactic molecular cloud conditions using RADEX)
        self.Texmin = 3.0  # K; (assume T_kin=5K, density = 1e3 cm^-3, column=1e13 cm^-2, sigv=3km/s)
        self.Texmax = 40  # K; T_k for Orion A is < 35 K. (T_k=40 at 1e5 cm^-3, 1e15 cm^-2, and 0.1 km/s --> Tex=37K)
        self.sigmin = 0.07  # km/s
        self.sigmax = 2.5  # km/s; (Larson's law for a 10pc cloud has sigma = 2.6 km/s)
        self.taumax = 30.0  # thick enough for NH3 (1,1) satellite hyperfines to have the same intensity as the main
        self.taumin = 0.1  # (note: NH3 (1,1) at 1e3 cm^-3, 1e13 cm^-2, 1 km/s linewidth, 40 K -> 0.15)
        self.eps = 0.001  # a small perturbation that can be used in guesses

        # a placeholder for if the window becomes line specific
        self.main_hf_moments = momgue.window_moments

        # for the NH3 (1,1) multicomponent model
        if self.fittype == 'nh3_multi_v':
            from pyspeckit.spectrum.models.ammonia_constants import freq_dict, voff_lines_dict, tau_wts_dict
            from . import ammonia_multiv as ammv
            from importlib import reload
            reload(ammv)

            self.linetype = 'nh3'
            # the current implementation only fits the 1-1 lines
            self.linename = 'oneone'
            self.linenames = [self.linename]
            self.fitter = ammv.nh3_multi_v_model_generator(n_comp=ncomp, linenames=self.linenames)

            self.freq_dict = freq_dict
            self.rest_value = freq_dict[self.linename] * u.Hz
            self.voffs = voff_lines_dict[self.linename]
            self.tau_wts = tau_wts_dict[self.linename]
            self.voff_at_peak = self.voffs[np.argmax(self.tau_wts)] # velocity offset at the brightest hyperfine line

            # the function to compute the spectral model
            self.model_func = ammv.ammonia_multi_v


        # for the N2H+ (1-0) multi-component model
        elif self.fittype == 'n2hp_multi_v':
            from .n2hp_constants import freq_dict, voff_lines_dict, tau_wts_dict
            from . import n2hp_multiv as n2hpmv

            self.linetype = 'n2hp'
            # the current implementation only fits the 1-0 lines
            self.linename = 'onezero'
            self.linenames = [self.linename]
            self.fitter = n2hpmv.n2hp_multi_v_model_generator(n_comp=ncomp, linenames=self.linenames)

            self.freq_dict = freq_dict
            self.rest_value = freq_dict[self.linename] * u.Hz
            self.voffs = voff_lines_dict[self.linename]
            self.tau_wts = tau_wts_dict[self.linename]
            self.voff_at_peak = self.voffs[np.argmax(self.tau_wts)] # velocity offset at the brightest hyperfine line

            # the function to compute the spectral model
            self.model_func = n2hpmv.n2hp_multi_v

            # change the parameter limits from the default to better reflects N2H+ (1-0)
            self.taumax = 40.0  # when the satellite hyperfine lines becomes optically thick

            # overwrite default windows, since N2H+ (1-0) have a satellite line that could be fairly bright
            self.window_hwidth = 4
            self.central_win_hwidth = 3.5

        else:
            raise FitTypeError("\'{}\' is an invalid fittype".format(fittype))

    def model_function(self, xarr, parameters, planemask=None, multithreaded=False):
        """
        Compute the spectral model for a given spectral axis and model parameters.

        Parameters
        ----------
        xarr : pyspeckit.spectrum.units.SpectroscopicAxis
            The spectral axis along which to evaluate the model function.
        parameters : array-like
            The model parameters for evaluation. Can be:

            - A 1D array with shape `(l,)`, representing a single set of parameters.
            - A 2D array with shape `(l, n)`, where `l` is the number of parameters,
              and `n` is the number of spatial positions.
            - A 3D array with shape `(l, m, n)`, where `l` is the number of parameters
              per pixel, and `m` and `n` represent the spatial dimensions. In this
              case, the function will evaluate the model at each `(m, n)` pixel
              specified by `planemask`.
        planemask : array-like, optional
            A 2D mask array with shape `(m, n)` indicating the pixels to evaluate when
            `parameters` is 3D. Pixels with `True` in the mask will be evaluated. If
            None, a default mask is generated where all non-zero, finite values in
            `parameters` are considered valid.
        multithreaded : bool, optional
            Whether to enable multithreading to evaluate pixels in parallel when
            `parameters` is 2D or 3D. Default is False.

        Returns
        -------
        np.ndarray
            The evaluated spectral model values. The output shape matches the input
            shape of `parameters`:

            - If `parameters` is 1D, returns a 1D array of computed values.
            - If `parameters` is 2D, returns a 1D array with length `n`, containing
              computed values for each position.
            - If `parameters` is 3D, returns a 2D array with shape `(m, n)`, containing
              computed values for each evaluated pixel in the masked region. Masked
              pixels are filled with NaN.
        """
        if parameters.ndim == 1:
            # Direct evaluation for 1D input
            return self.model_func(xarr, *parameters)

        elif parameters.ndim == 2:
            # For 2D input, transpose for easier iteration
            pars = parameters.T

            # Initialize results array for 2D output
            results = np.full(pars.shape[0], np.nan, dtype=object)

            # Define evaluation function for each position
            def evaluate_position(i):
                par = pars[i]
                return self.model_func(xarr, *par)

            if multithreaded:
                with ThreadPoolExecutor() as executor:
                    evaluated_results = list(executor.map(evaluate_position, range(len(pars))))
                results[:] = evaluated_results
            else:
                for i, par in enumerate(pars):
                    results[i] = self.model_func(xarr, *par)

            return results

        elif parameters.ndim == 3:
            # Generate mask if not provided
            if planemask is None:
                planemask = np.all(np.isfinite(parameters), axis=0) & np.all(parameters != 0, axis=0)

            # Mask parameters for efficient evaluation
            pars = parameters[:, planemask].T

            # Initialize results array to hold model outputs for 3D input
            results = np.full((parameters.shape[1], parameters.shape[2]), np.nan, dtype=object)

            # Define evaluation function for each pixel
            def evaluate_pixel(i):
                par = pars[i]
                return self.model_func(xarr, *par)

            if multithreaded:
                with ThreadPoolExecutor() as executor:
                    evaluated_results = list(executor.map(evaluate_pixel, range(len(pars))))
                results[planemask] = evaluated_results
            else:
                for i, par in enumerate(pars):
                    results[planemask][i] = self.model_func(xarr, *par)

            return results

    def peakT(self, parameters, index_v=0, planemask=None, multithreaded=False):
        """
        Estimate the peak brightness temperature of the spectral model by evaluating
        the model at the spectral location (i.e., the velocity) of the brightest
        hyperfine line.

        Note
        ----
        This method only works for a single-component model, where the peak
        brightness is expected to be located near the brightest hyperfine line.

        Parameters
        ----------
        parameters : array-like
            The model parameters to evaluate, where each element corresponds to a
            parameter used in the spectral model. Can be:

            - A 1D array with shape `(l,)`, representing a single set of parameters.
            - A 2D array with shape `(l, n)`, where `l` is the number of parameters and
              `n` is the number of spatial positions.
            - A 3D array with shape `(l, m, n)`, where `l` is the number of parameters
              per pixel, and `m` and `n` represent the spatial dimensions.
        index_v : int, optional
            Index of the parameter corresponding to the velocity centroid, which is 0
            for all spectral models implemented in MUFASA. Default is 0.
        planemask : array-like, optional
            A 2D mask array with shape `(m, n)` indicating the pixels to evaluate when
            `parameters` is 3D. Pixels with `True` in the mask will be evaluated. If
            None, a mask is generated based on finite, non-zero values in
            `parameters`.
        multithreaded : bool, optional
            Whether to enable multithreading to evaluate pixels in parallel when
            `parameters` is 2D or 3D. Default is False.

        Returns
        -------
        np.ndarray
            An array with the peak brightness temperature for each evaluated pixel:

            - If `parameters` is 1D, returns a scalar representing the peak
              brightness temperature.
            - If `parameters` is 2D, returns a 1D array with length `n`, containing
              the peak brightness temperature for each position.
            - If `parameters` is 3D, returns a 2D array with shape `(m, n)`, containing
              the peak brightness temperature for each evaluated pixel in the masked
              region. Masked pixels are filled with NaN.
        """
        # Function implementation

        # Get the velocity at the brightest hyperfine line
        v0 = np.array([parameters[index_v] + self.voff_at_peak]) * u.km / u.s

        # Convert the velocity to frequency for evaluation
        nu0 = v0.to(u.GHz, equivalencies=u.doppler_radio(self.rest_value))

        if parameters.ndim == 1:
            # Direct evaluation for 1D input
            return self.model_func(nu0, *parameters)[0]

        else:
            if parameters.ndim == 2:
                pars = parameters.T
                nu0_array = nu0.T

            if parameters.ndim == 3:
                # Generate mask if not provided
                if planemask is None:
                    planemask = np.all(np.isfinite(parameters), axis=0) & np.all(parameters != 0, axis=0)

                # Mask parameters and frequencies for efficient evaluation
                pars = parameters[:, planemask].T
                nu0_array = nu0[:, planemask].T

            # Initialize results array
            pT = np.full(nu0_array.shape[0], np.nan)

            # Define a function for multithreading
            def evaluate_pixel(i):
                nu0_i = nu0_array[i]
                par = pars[i]
                return self.model_func(nu0_i, *par)[0]

            if multithreaded:
                # Use ThreadPoolExecutor for parallel processing
                with ThreadPoolExecutor() as executor:
                    results = list(executor.map(evaluate_pixel, range(len(pars))))
                pT[:] = results
            else:
                # Single-threaded loop if multicore is False
                for i, par in enumerate(pars):
                    nu0_i = nu0_array[i]
                    pT[i] = self.model_func(nu0_i, *par)[0]

            if parameters.ndim == 3:
                # Reconstruct the 2D output
                pT2D = np.full(planemask.shape, np.nan)
                pT2D[planemask] = pT
                return pT2D

            else:
                return pT



__author__ = 'mcychen'


"""
==================================================================
Ammonia inversion transition: deblended fitter (Hyperfine-removed)
==================================================================
"""

#=======================================================================================================================

from pyspeckit.spectrum.models import hyperfine
from .n2hp_constants import (line_names, freq_dict)
#=======================================================================================================================


tau0 = 1.0

# represent the tau profile of nh3 spectra as a single Gaussian for each individual velocity slab
n2hp_vtau_deblended = {linename:
            hyperfine.hyperfinemodel({0:0},
                                     {0:0.0},
                                     {0:freq_dict[linename]},
                                     {0:tau0},
                                     {0:1},
                                    )
            for linename in line_names}


def n2hp_vtau_singlemodel_deblended(xarr, Tex, tau, xoff_v, width, linename = 'onezero'):
    # the parameters are in the order of vel, width, tex, tau for each velocity component
    return n2hp_vtau_deblended[linename].hyperfine(xarr, Tex=Tex, tau=tau, xoff_v=xoff_v, width=width)



"""
The `mufasa.signals` module provides robust tools for preprocessing spectral data, including RMS estimation,
signal masking, spatial trimming, and moment calculations.
"""

__author__ = "Mike Chen"

import gc

#=======================================================================================================================

import numpy as np
from spectral_cube.masks import LazyComparisonMask
from skimage.morphology import binary_dilation, remove_small_objects, binary_erosion, disk, binary_opening, ball
import astropy.units as u

#=======================================================================================================================
# Import in-house functions
from .utils.mufasa_log import get_logger
logger = get_logger(__name__)
#=======================================================================================================================


def get_snr(cube, rms=None, **kwargs):
    """
    Calculate the peak signal-to-noise ratio of the cube.

    Parameters
    ----------
    cube : SpectralCube
        SpectralCube object.
    rms : float or None, optional
        Root mean square of the map. If None, calculate using `get_rms_robust()`.
    **kwargs
        Additional arguments for RMS calculation.

    Returns
    -------
    Quantity
        Peak signal-to-noise ratio.
    """
    if rms is None:
        rms = get_rms_robust(cube, **kwargs)

    return cube.max(axis=0) / rms #, how='ray'


def get_rms_robust(cube, sigma_cut=2.5, expand=20, trim=None, method='robust', return_sigmask=False, **kwargs):
    """
    Make a robust RMS estimate.

    Parameters
    ----------
    cube : SpectralCube
        SpectralCube object.
    sigma_cut : float, optional
        Signal-to-noise ratio to mask out regions containing signal.
    expand : int, optional
        Number of spectral channels to expand the signal mask.
    trim : int or None, optional
        Number of pixels to trim from the edges of the cube.
    method : {'robust', 'signal_mask'}, optional
        Method to calculate RMS.
    return_sigmask : bool, optional
        Whether to return the 3D signal mask along with the RMS.
    **kwargs
        Additional arguments.

    Returns
    -------
    Quantity or tuple of Quantity
        RMS value, and optionally the signal mask.
    """
    if trim is not None:
        cube = trim_cube_edge(cube, trim)

    if method == 'robust':
        rms = cube.mad_std(axis=0) #how='ray'
        # Refine RMS by masking signal regions
        rms, sig_mask = refine_rms(cube, rms, sigma_cut=sigma_cut, expand=expand)

    elif method == 'signal_mask':
        # Use signal mask method to calculate RMS
        sig_mask = get_signal_mask(cube, sigma_cut=sigma_cut, **kwargs)
        noise_cube = cube.with_mask(~sig_mask & cube.mask.include())
        rms = noise_cube.mad_std(axis=0) #how='ray'

    if return_sigmask:
        return rms, sig_mask
    else:
        return rms


def refine_rms(cube, rms, sigma_cut=3, expand=20):
    """
    Refine the RMS estimate by masking out signal regions.

    Parameters
    ----------
    cube : SpectralCube
        SpectralCube object.
    rms : Quantity
        Initial RMS estimate.
    sigma_cut : float, optional
        Signal-to-noise threshold to mask signals.
    expand : int, optional
        Number of spectral channels to expand the signal mask.

    Returns
    -------
    Quantity
        Refined RMS.
    ndarray
        Signal mask.
    """
    # Mask above the cut threshold
    sig_mask = (cube > sigma_cut * rms)
    sig_mask = binary_dilation(sig_mask.include(), np.ones(shape=(expand, 1, 1), dtype=bool))
    noise_cube = cube.with_mask(~sig_mask & cube.mask.include())
    rms = noise_cube.mad_std(axis=0) #how='ray'
    return rms, sig_mask


def get_signal_mask(cube, snr_cut=3, minsize=2, expand=20, clean=True, return_rms=False):
    """
    Provide a 3D mask indicating signal regions based on RMS and SNR threshold.

    Parameters
    ----------
    cube : SpectralCube
        SpectralCube object.
    snr_cut : float, optional
        Signal-to-noise threshold to define signals.
    minsize : int, optional
        Minimum size of connected components in spectral direction.
    expand : int, optional
        Number of spectral channels to expand signal regions.
    clean : bool, optional
        Remove spatially isolated pixels.
    return_rms : bool, optional
        Whether to return RMS along with the signal mask.

    Returns
    -------
    ndarray
        Signal mask.
    Quantity, optional
        RMS value if `return_rms` is True.
    """
    rms = cube.mad_std(axis=0) #how='ray'
    # Mask above the cut threshold
    sig_mask = (cube > snr_cut * rms)
    sig_mask = refine_signal_mask(sig_mask, minsize=minsize, expand=expand, clean=clean)

    if return_rms:
        return sig_mask, rms
    else:
        return sig_mask


def refine_signal_mask(sig_mask, minsize=2, expand=20, clean=True):
    """
    Refine a signal mask by removing noisy features and expanding the mask.

    Parameters
    ----------
    sig_mask : LazyComparisonMask or ndarray
        Initial signal mask.
    minsize : int, optional
        Minimum number of continuous spectral channels for signal mask.
    expand : int, optional
        Number of spectral channels to expand signal regions.
    clean : bool, optional
        Remove spatially isolated pixels.

    Returns
    -------
    ndarray
        Refined signal mask.
    """
    if isinstance(sig_mask, LazyComparisonMask):
        sig_mask = sig_mask.include()

    # Remove voxels that are likely just statistical noise
    sig_mask = binary_erosion(sig_mask, np.ones(shape=(minsize, 1, 1), dtype=bool))
    sig_mask = remove_small_objects(sig_mask, 1)
    # Expand the "signal" regions a bit
    struct = np.ones(shape=(expand, 1, 1), dtype=bool)
    sig_mask = binary_dilation(sig_mask, struct)
    sig_mask = binary_dilation(sig_mask, disk(3)[np.newaxis, :])

    if clean:
        # Remove isolated pixels
        sig_mask *= remove_small_objects(binary_erosion(np.any(sig_mask, axis=0)), 1)

    return sig_mask


def trim_cube_edge(cube, trim=3, min_size=1000):
    """
    Remove spatial edges from a cube.

    Parameters
    ----------
    cube : SpectralCube
        SpectralCube object.
    trim : int, optional
        Number of pixels to trim from edges.
    min_size : int, optional
        Minimum size of cube for trimming to be applied.

    Returns
    -------
    SpectralCube
        Trimmed cube.
    """
    mask = cube.mask.include()
    planemask = np.any(mask, axis=0)

    if np.sum(planemask) > min_size:
        # Trim edges using a 2D mask
        planemask = trim_edge(planemask, trim=trim)
        mask[:, ~planemask] = False
        return cube.with_mask(mask)
    else:
        return cube


def trim_edge(mask, trim=3):
    """
    Trim edges using a 2D mask.

    Parameters
    ----------
    mask : ndarray
        2D mask.
    trim : int, optional
        Number of pixels to trim.

    Returns
    -------
    ndarray
        Trimmed mask.
    """
    # Set all edge pixels to False
    mask[[0, -1], :] = False
    mask[:, [0, -1]] = False
    return binary_erosion(mask, disk(trim))


def v_estimate(cube, rms, snr_cut=3):
    """
    Estimate the velocity centroid based on peak emission.

    Parameters
    ----------
    cube : SpectralCube
        SpectralCube object.
    rms : Quantity
        Root mean square of the map.
    snr_cut : float, optional
        Signal-to-noise threshold for quality control.

    Returns
    -------
    Quantity
        Velocity centroid estimate.
    """
    sig_mask = cube > rms * snr_cut
    # Clean the mask to remove noisy features
    sig_mask = binary_erosion(sig_mask.include(), ball(2))
    return get_v_at_peak(cube.with_mask(sig_mask))


def get_v_at_peak(cube_masked):
    """
    Find the velocity corresponding to the peak emission.

    Parameters
    ----------
    cube_masked : SpectralCube
        SpectralCube object with mask.

    Returns
    -------
    ndarray
        Velocity at peak emission.
    """
    idx_max = cube_masked.argmax(axis=0)#, how='slice')
    v_est = cube_masked.spectral_axis[idx_max]
    # Set velocity to NaN where no valid data is present
    v_est[~np.any(cube_masked.get_mask_array(), axis=0)] = np.nan
    return v_est.value


def get_v_mask(cube_masked, v_center, rms=None, window_hwidth=5):
    """
    Return a mask centered on a reference velocity with a spectral window.

    Parameters
    ----------
    cube_masked : SpectralCube
        SpectralCube object with mask.
    v_center : Quantity
        Reference velocity at center of spectral window.
    rms : Quantity or None, optional
        Root mean square of the map.
    window_hwidth : float, optional
        Half-width of spectral window in km/s.

    Returns
    -------
    ndarray
        Velocity mask.
    """
    snr_cut = 3
    # Get window centered on v_center with a half-width of window_hwidth
    window_mask = (cube_masked.spectral_axis[:, np.newaxis, np.newaxis] < (v_center + window_hwidth) * u.km / u.s) & \
                  (cube_masked.spectral_axis[:, np.newaxis, np.newaxis] > (v_center - window_hwidth) * u.km / u.s)

    if rms is not None:
        # Remove voxels that are likely noise
        mask_new = binary_opening((cube_masked > rms * snr_cut).include())
        return mask_new & window_mask
    else:
        return cube_masked.get_mask_array() & window_mask


def get_moments(cube, window_hwidth=5, linewidth_sigma=True, trim=3, return_rms=False, central_win_hwidth=None,
                **kwargs):
    """
    Calculate moments of the signals in a cube.

    Parameters
    ----------
    cube : SpectralCube
        SpectralCube object.
    window_hwidth : float, optional
        Half-width of spectral window for moment calculation.
    linewidth_sigma : bool, optional
        Whether to calculate linewidth sigma instead of moment 2 (see spectral_cube documentation for further details).
    trim : int, optional
        Number of pixels to trim from edges of the cube.
    return_rms : bool, optional
        Whether to return RMS along with moments.
    central_win_hwidth : float, optional
        If provided, re-evulate velocity at peak using a window centered on the mode of the initial velocity estimate, with a
        half-wdith of central_win_hwidth. Note that window is the same for all pixels, but only applied as an intermediate step.
        This should only used if hyperfine lines far from the main hyperfines have a comprable brightness.

    **kwargs
        Additional arguments.

    Returns
    -------
    tuple of Quantity
        Tuple of moment maps (moment 0, moment 1, linewidth or moment 2, optionally RMS).
    """
    snr_cut = 3
    noise_mask_cut = 2.5

    # Trim cube edges
    if trim is not None:
        cube = trim_cube_edge(cube, trim=trim)
    rms = get_rms_robust(cube, sigma_cut=noise_mask_cut, **kwargs)
    v_peak = v_estimate(cube, rms, snr_cut=snr_cut)

    if central_win_hwidth is not None:
        # impose a strick window based on the esitmated mode velocity to re-evaulate v_peak
        v_mode = estimate_mode(v_peak, bins=25)
        signal_mask_tmp = get_v_mask(cube, v_center=v_mode, rms=rms, window_hwidth=central_win_hwidth)
        cube_signal_tmp = cube.with_mask(signal_mask_tmp)
        v_peak = v_estimate(cube_signal_tmp, rms, snr_cut=snr_cut)
        del signal_mask_tmp
        del cube_signal_tmp
        gc.collect()

    # Get velocity mask centered on v_peak with window_hwidth
    signal_mask = get_v_mask(cube, v_peak, rms=rms, window_hwidth=window_hwidth)
    # Expand the footprint a bit
    signal_mask = binary_dilation(signal_mask, disk(3)[np.newaxis, :])

    cube_signal = cube.with_mask(signal_mask)
    mom0 = cube_signal.moment0()
    mom1 = cube_signal.moment1()
    output = (mom0, mom1)

    if linewidth_sigma:
        output += (cube_signal.linewidth_sigma(),)
    else:
        output += (cube_signal.moment2(),)

    if return_rms:
        output += (rms,)

    return output


def estimate_mode(data, bins=25):
    """
    Estimate the mode of the data using a histogram.

    Parameters
    ----------
    data : ndarray
        The data for which to estimate the mode. Non-finite values will be ignored.
    bins : int, optional
        Number of bins to use for the histogram. Default is 25.

    Returns
    -------
    float
        The estimated mode of the data.
    """
    data=data[np.isfinite(data)]
    counts, bins = np.histogram(data, bins=bins)

    # Find the bin with the maximum frequency
    max_bin_index = np.argmax(counts)
    return (bins[max_bin_index] + bins[max_bin_index + 1]) / 2
